import logging
import os
import sys
import click
import logging
import json
import csv

from pathlib import Path
from typing import DefaultDict, Iterator, List, TypeVar

from mzspeclib.spectrum_library import SpectrumLibrary
from mzspeclib.index import MemoryIndex, SQLIndex
from mzspeclib.backends.base import FormatInferenceFailure, SpectralLibraryBackendBase
from mzspeclib.backends.utils import PreBufferedStreamReader
from mzspeclib.validate import validator
from mzspeclib.validate.level import RequirementLevel
from mzspeclib.ontology import ControlledVocabularyResolver

from mzspeclib.tools.utils import ColoringFormatter

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

logger = logging.getLogger(__name__)


def _input_format_option():
    return click.option(
        "-i",
        "--input-format",
        type=click.Choice(sorted(SpectralLibraryBackendBase._file_extension_to_implementation)),
        help="The file format of the input file. If omitted, will attempt to infer automatically.",
    )


def _display_tree(tree, indent: int=0):
    if isinstance(tree, dict):
        if not tree:
            return
        for k, v in tree.items():
            if isinstance(v, (list, dict)) and not v:
                continue
            click.echo(f"{' ' * indent}└-{k}", err=True, nl=True)
            _display_tree(v, indent + 2)
    elif isinstance(tree, list):
        for v in tree:
            _display_tree(v, indent + 2)
    else:
        click.echo(f"{' ' * indent}└-{tree}", err=True, nl=True)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("-d", "--debug-logging", is_flag=True, help="Enable debug logging")
@click.option("-l", "--log-file", type=click.Path(writable=True, path_type=Path), help="Write log messages to this file as well as STDERR")
def main(debug_logging=False, log_file: Path=None):
    """A collection of utilities for inspecting and manipulating spectral libraries."""
    format_string = '[%(asctime)s] %(levelname).1s | %(name)s | %(filename)s:%(funcName)s:%(lineno)d | %(message)s'

    logging.basicConfig(
        level='INFO' if not debug_logging else "DEBUG",
        stream=sys.stderr,
        format=format_string,
        datefmt="%H:%M:%S")

    fmtr = ColoringFormatter(format_string, datefmt='%H:%M:%S')

    for handler in logging.getLogger().handlers:
        handler.setFormatter(
            fmtr
        )

    if log_file is not None:
        if not log_file.parent.exists():
            os.makedirs(log_file.parent)
        handler = logging.FileHandler(str(log_file), mode='w')
        handler.setLevel(logging.INFO if not debug_logging else logging.DEBUG)
        handler.setFormatter(
            logging.Formatter(format_string, "%H:%M:%S")
        )
        logging.getLogger().addHandler(handler)

    if debug_logging:
        sys.excepthook = _debug_hook


@main.command("describe", short_help=("Produce a minimal textual description"
                                      " of a spectral library"))
@click.argument('path', type=click.Path(exists=True))
# @click.option("-d", "--diagnostics", is_flag=True,
#               help="Run more diagnostics, greatly increasing runtime but producing additional information")
@_input_format_option()
def describe(path, diagnostics=False, input_format=None):
    """Produce a minimal textual description of a spectral library."""
    click.echo("Describing \"%s\"" % (path,))
    if SQLIndex.exists(path):
        index_type = SQLIndex
    else:
        index_type = MemoryIndex
    try:
        library = SpectrumLibrary(filename=path, index_type=index_type, format=input_format)
    except FormatInferenceFailure as err:
        click.echo(f"{err}", err=True)
        raise click.Abort()
    click.echo(f"Format: {library.format}")
    click.echo(f"Spectrum Count: {library.__len__()}")
    for attr in library.attributes:
        if not attr.group_id:
            click.echo(f"{attr.key}={attr.value}")
        else:
            click.echo(f"[{attr.group_id}]{attr.key}={attr.value}")


@main.command("convert", short_help=("Convert a spectral library from one format to another"))
@click.argument('inpath', type=click.Path(exists=True, allow_dash=False))
@click.argument("outpath", type=click.Path())
@_input_format_option()
@click.option("-f", "--format", type=click.Choice(["text", "json", "msp"]), default="text")
@click.option("-k", "--library-attribute", "library_attributes", type=(str, str), multiple=True,
              help="Specify an attribute to add to the library metadata section. May be repeated.")
@click.option("-K", "--header-file", type=click.Path(readable=True),
              help="Specify a file to read name-value pairs from. May be JSON or TAB-separated")
def convert(inpath, outpath, format=None, header_file=None, library_attributes=(), input_format=None):
    """
    Convert a spectral library from one format to another. If `outpath` is `-`,
    instead of writing to file, data will instead be sent to STDOUT.
    """
    if format is None:
        format = "text"
    if format == 'msp':
        click.secho("MSP conversion is arbitrary and lossy", fg='yellow', err=True)
    if SQLIndex.exists(inpath):
        index_type = SQLIndex
    else:
        index_type = MemoryIndex
    click.echo(f"Opening {inpath}", err=True)

    create_index = True
    # if inpath == '-':
    #     inpath = PreBufferedStreamReader(click.get_binary_stream("stdin"))
    #     create_index = False
    try:
        library = SpectrumLibrary(filename=inpath, index_type=index_type, format=input_format, create_index=create_index)
    except FormatInferenceFailure as err:
        click.echo(f"{err}", err=True)
        raise click.Abort()
    if header_file:
        library_attributes = list(library_attributes)
        if header_file.endswith(".json"):
            library_attributes.extend(
                json.load(open(header_file, 'rt')).items())
        else:
            library_attributes.extend(
                csv.reader(open(header_file, 'rt'), delimiter='\t'))
    if library_attributes:
        resolver = ControlledVocabularyResolver()
        for k, v in library_attributes:
            k = resolver.attribute_syntax(k)
            library.add_attribute(k, v)
    click.echo(f"Writing to {outpath}", err=True)
    fh = click.open_file(outpath, mode='w')
    try:
        library.write(fh, format)
    except IOError as err:
        if outpath == '-':
            return
        else:
            raise


    _display_tree(library.summarize_parsing_errors())


@main.command("index", short_help="Build an on-disk index for a spectral library")
@click.argument('inpath', type=click.Path(exists=True))
@_input_format_option()
def build_index(inpath, input_format=None):
    """Build an external on-disk SQL-based index for the spectral library"""
    try:
        library = SpectrumLibrary(filename=inpath, index_type=SQLIndex, format=input_format)
    except FormatInferenceFailure as err:
        click.echo(f"{err}", err=True)
        raise click.Abort()

T = TypeVar('T')

def _progress_logger(iterable: Iterator[T], label, increment: int=100) -> Iterator[T]:
    n = len(iterable)
    for i, item in enumerate(iterable):
        if i % increment == 0 and i:
            logger.info(f"... {label} {i}/{n} ({i * 100 / n:0.2f}%)", stacklevel=2)
        yield item


@main.command(short_help="Semantically validate a spectral library")
@click.argument('inpath', type=click.Path(exists=True))
@click.option("-p", "--profile", "profiles", type=click.Choice(
    ["consensus", "single", "silver", "peptide", "gold"],
    case_sensitive=False),
    multiple=True)
@_input_format_option()
def validate(inpath, profiles=None, input_format=None):
    """Semantically and structurally validate a spectral library."""
    if profiles is None:
        profiles = []
    if SQLIndex.exists(inpath):
        index_type = SQLIndex
    else:
        index_type = MemoryIndex

    logger.info(f"Loading library {inpath}...")
    try:
        library = SpectrumLibrary(filename=inpath, index_type=index_type, format=input_format)
    except FormatInferenceFailure as err:
        click.echo(f"{err}", err=True)
        raise click.Abort()

    logger.info(f"Loading validators...")
    chain = validator.load_default_validator()
    for profile in profiles:
        if profile is None:
            continue
        logger.info(f"... {profile}")
        chain |= validator.get_validator_for(profile)

    logger.info(f"Validating {inpath}...")
    n_spectra = len(library)
    increment = max(min(n_spectra // 10, 5000), 1)
    chain.validate_library(library, _progress_logger(library, "Validating spectra", increment))

    by_level: DefaultDict[RequirementLevel, List[validator.ValidationError]] = DefaultDict(list)
    for message in chain.error_log:
        by_level[message.requirement_level].append(message)

    for level, bucket in sorted(by_level.items()):
        log_level = logging.WARN
        if level == RequirementLevel.may:
            log_level = logging.DEBUG
        logger.log(log_level, f"Found {len(bucket)} violations for {level.name.upper()} rules")
        for err in bucket:
            logger.log(log_level, f"... {err.message}")


def _debug_hook(type, value, tb):
    if not sys.stderr.isatty():
        click.secho("Running interactively, not starting debugger", fg="yellow")
        sys.__excepthook__(type, value, tb)
    else:
        import pdb
        import traceback
        traceback.print_exception(type, value, tb)
        pdb.post_mortem(tb)


if __name__ == "__main__":
    main()



"""
Read EncyclopeDIA SQLite3 spectral libraries.

For more information, see `EncyclopeDIA's website <https://www.searlelab.org/software/encyclopedia/index.html>`_
"""
import sqlite3
import zlib
from dataclasses import dataclass

from typing import Any, Iterator, List, Mapping, Tuple, Iterable, Type

import numpy as np

from pyteomics import proforma

from mzspeclib import annotation
from mzspeclib.analyte import FIRST_ANALYTE_KEY, FIRST_INTERPRETATION_KEY, Analyte, ProteinDescription
from mzspeclib.spectrum import Spectrum, SPECTRUM_NAME, CHARGE_STATE
from mzspeclib.attributes import AttributeManager, Attributed, Attribute

from mzspeclib.backends.base import SpectralLibraryBackendBase, FORMAT_VERSION, DEFAULT_VERSION

from mzspeclib.index.base import IndexBase
from mzspeclib import const as c
from mzspeclib.const import CHARGE_STATE, DECOY_SPECTRUM, DECOY_PEPTIDE_SPECTRUM, PROFORMA_ION, STRIPPED_PEPTIDE_SEQ, LIBRARY_CREATION_SW


def _decode_peaks(record: sqlite3.Row):
    raw_data = zlib.decompress(record['MassArray'])
    mass_array = np.frombuffer(raw_data, dtype='>d')
    raw_data = zlib.decompress(record['IntensityArray'])
    intensity_array = np.frombuffer(raw_data, dtype='>f')
    return mass_array, intensity_array


@dataclass
class EncyclopediaIndexRecord:
    number: int
    precursor_mz: float
    precursor_charge: int
    peptide: str


class EncyclopediaIndex(IndexBase):
    connection: sqlite3.Connection

    def __init__(self, connection):
        self.connection = connection

    def __getitem__(self, i):
        if isinstance(i, int):
            return self.search(i + 1)
        elif isinstance(i, slice):
            return [self.search(j + 1) for j in range(i.start or 0, i.stop or len(self), i.step or 1)]
        else:
            raise TypeError(f"Cannot index {self.__class__.__name__} with {i}")

    def _record_from(self, row: Mapping) -> EncyclopediaIndexRecord:
        peptide_sequence = row['PeptideModSeq']
        return EncyclopediaIndexRecord(row['rowid'], row['PrecursorMz'], row['PrecursorCharge'], peptide_sequence)

    def search(self, i):
        if isinstance(i, int):
            info = self.connection.execute("SELECT rowid, PrecursorMz, PrecursorCharge, PeptideModSeq FROM entries WHERE rowid = ?", (i, )).fetchone()
            return self._record_from(info)
        elif isinstance(i, str):
            raise NotImplementedError()

    def __iter__(self):
        return map(self._record_from, self.connection.execute("SELECT rowid, PrecursorMz, PrecursorCharge, PeptideModSeq FROM entries ORDER BY rowid").fetchall())

    def __len__(self):
        return self.connection.execute("SELECT count(rowid) FROM entries;").fetchone()[0]


class EncyclopediaSpectralLibrary(SpectralLibraryBackendBase):
    """Read EncyclopeDIA SQLite3 spectral library files."""

    connection: sqlite3.Connection

    file_format = "dlib"
    format_name = "encyclopedia"

    @classmethod
    def has_index_preference(cls, filename) -> Type[IndexBase]:
        return EncyclopediaIndex

    def __init__(self, filename: str, **kwargs):
        super().__init__(filename)
        self.connection = sqlite3.connect(filename)
        self.connection.row_factory = sqlite3.Row
        self.index = EncyclopediaIndex(self.connection)
        self.read_header()

    def read_header(self) -> bool:
        attribs = AttributeManager()
        attribs.add_attribute(FORMAT_VERSION, DEFAULT_VERSION)
        attribs.add_attribute(LIBRARY_CREATION_SW, "EncyclopeDIA")
        self.attributes = attribs
        return True

    def _populate_analyte(self, analyte: Analyte, row: Mapping[str, Any]):
        """
        Fill an analyte with details describing a peptide sequence and inferring
        from context its traits based upon the assumptions EncyclopeDIA makes.

        EncyclopeDIA only stores modifications as delta masses.
        """
        peptide = proforma.ProForma.parse(row['PeptideModSeq'])
        peptide.charge_state = row['PrecursorCharge']
        analyte.add_attribute(PROFORMA_ION, str(peptide))
        analyte.add_attribute(c.THEORETICAL_MASS, peptide.mass)
        analyte.add_attribute(STRIPPED_PEPTIDE_SEQ, row['PeptideSeq'])

        cursor = self.connection.execute(
            "SELECT ProteinAccession, isDecoy FROM peptidetoprotein WHERE PeptideSeq = ?;", (row['PeptideSeq'], ))

        had_decoy = False
        for protrow in cursor:
            accession = protrow['ProteinAccession']
            is_decoy = bool(int(protrow['isDecoy']))
            had_decoy = had_decoy or is_decoy
            analyte.add_attribute_group([
                Attribute('MS:1000885|protein accession', accession)
            ])
        return had_decoy

    def get_spectrum(self, spectrum_number: int = None, spectrum_name: str = None):
        """
        Read a spectrum from the spectrum library.

        EncyclopeDIA does not support alternative labeling of spectra with a
        plain text name so looking up by `spectrum_name` is not supported.
        """
        if spectrum_number is None:
            raise ValueError("Only spectrum number queries are supported. spectrum_number must have an integer value")
        try:
            spectrum_number = int(spectrum_number)
        except (TypeError, ValueError):
            raise ValueError(f"spectrum_number must have an integer value, received {spectrum_number!r}") from None

        info = self.connection.execute("SELECT rowid, * FROM entries WHERE rowid = ?;", (spectrum_number, )).fetchone()
        spectrum = self._new_spectrum()
        spectrum.key = info['rowid']
        spectrum.index = info['rowid'] - 1
        spectrum.precursor_mz = info['PrecursorMz']
        spectrum.add_attribute(CHARGE_STATE, info["PrecursorCharge"])
        try:
            spectrum.add_attribute(c.RETENTION_TIME, info['RTInSeconds'] / 60.0)
        except KeyError:
            pass

        try:
            spectrum.add_attribute(
                c.SOURCE_FILE,
                f"file://{info['SourceFile']}"
            )
        except KeyError:
            pass


        analyte = self._new_analyte(1)
        had_decoy = self._populate_analyte(analyte, info)
        if had_decoy:
            spectrum.add_attribute(DECOY_SPECTRUM, DECOY_PEPTIDE_SPECTRUM)

        spectrum.add_analyte(analyte)

        interp = self._new_interpretation(1)
        interp.add_analyte(analyte)
        spectrum.add_interpretation(interp)

        mz_array, intensity_array = _decode_peaks(info)
        n_peaks = len(mz_array)
        spectrum.add_attribute(c.NUM_PEAKS, n_peaks)

        peak_list = []
        # EncyclopeDIA does not encode product ion identities
        for i, mz in enumerate(mz_array):
            row = (mz, intensity_array[i], [], [])
            peak_list.append(row)
        spectrum.peak_list = peak_list
        return spectrum

    def read(self) -> Iterator[Spectrum]:
        for rec in self.index:
            yield self.get_spectrum(rec.number)



"""Read mzSpecLib Text Format"""
import re
import os
import io
import logging
import warnings
import enum
import numbers

from dataclasses import dataclass
from collections import deque
from typing import ClassVar, List, Optional, Tuple, Union, Iterable

from mzspeclib.annotation import parse_annotation
from mzspeclib.spectrum import Spectrum
from mzspeclib.cluster import SpectrumCluster
from mzspeclib.attributes import Attribute, AttributeManager, AttributeSetRef, Attributed, AttributeSet
from mzspeclib.analyte import Analyte, Interpretation, InterpretationMember
from mzspeclib.utils import ValidationWarning

from .base import (
    DEFAULT_VERSION,
    SpectralLibraryBackendBase,
    _PlainTextSpectralLibraryBackendBase,
    SpectralLibraryWriterBase,
    FORMAT_VERSION,
    AttributeSetTypes,
)
from .utils import try_cast, open_stream

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


term_pattern = re.compile(
    r"^(?P<term>(?P<term_accession>\S+:(?:\d|X)+)\|(?P<term_name>[^=]+))"
)
key_value_term_pattern = re.compile(
    r"^(?P<term>(?P<term_accession>[A-Za-z0-9:.]+:(?:\d|X)+)\|(?P<term_name>[^=]+?))\s*=(\s*(?P<value>.+))?"
)
grouped_key_value_term_pattern = re.compile(
    r"^\[(?P<group_id>\d+)\](?P<term>(?P<term_accession>\S+:(?:\d|X)+)\|(?P<term_name>[^=]+?))\s*=(\s*(?P<value>.+))?"
)
float_number = re.compile(r"^\d+(.\d+)?")


class _SpectrumParserStateEnum(enum.Enum):
    unknown = 0
    header = 1
    analyte = 2
    interpretation = 3
    interpretation_member = 4
    peaks = 5
    done = 6
    cluster = 7


class _LibraryParserStateEnum(enum.Enum):
    unknown = 0
    header = 1
    attribute_sets = 2
    content = 3


ATTRIBUTE_SET_NAME = "MS:1003212|library attribute set name"
PEAK_ATTRIBUTE = "MS:1003254|peak attribute"

START_OF_SPECTRUM_MARKER = re.compile(r"^<(?:Spectrum)(?:\s*=\s*(.+))?>")
START_OF_INTERPRETATION_MARKER = re.compile(r"^<Interpretation(?:\s*=\s*(.+))>")
START_OF_ANALYTE_MARKER = re.compile(r"^<Analyte(?:\s*=\s*(.+))>")
START_OF_PEAKS_MARKER = re.compile(r"^<Peaks>")
START_OF_LIBRARY_MARKER = re.compile(r"^<mzSpecLib\s*(.+)?>")
START_OF_INTERPRETATION_MEMBER_MARKER = re.compile(
    r"<InterpretationMember(?:\s*=\s*(.+))>"
)
START_OF_ATTRIBUTE_SET = re.compile(
    r"<AttributeSet (Spectrum|Analyte|Interpretation|Cluster)\s*=\s*(.+)>"
)
START_OF_CLUSTER = re.compile(r"<Cluster(?:\s*=\s*(.+))>")

SPECTRUM_NAME_PRESENT = re.compile(r"MS:1003061\|(?:library )?spectrum name\s*=\s*")
SPECTRUM_NAME_MATCH = re.compile(r"MS:1003061\|(?:library )?spectrum name\s*=\s*(.+)")

FALLBACK_PEAK_LINE_PATTERN = re.compile(
    r"(?P<mz>\d+(?:\.\d+)?)\s+(?P<intensity>\d+(?:\.\d+)?)(?:\s+(?P<rest>.+))?"
)

attribute_set_types = {
    "spectrum": AttributeSetTypes.spectrum,
    "analyte": AttributeSetTypes.analyte,
    "interpretation": AttributeSetTypes.interpretation,
    "cluster": AttributeSetTypes.cluster,
}


class _Scope:
    state: _SpectrumParserStateEnum
    attribute_group: Optional[str]
    working_attribute_group: Optional[str]

    def __init__(
        self,
        state: _SpectrumParserStateEnum,
        attribute_group: Optional[str] = None,
        working_attribute_group: Optional[str] = None,
    ) -> None:
        if working_attribute_group is None:
            working_attribute_group = attribute_group
        self.state = state
        self.attribute_group = attribute_group
        self.working_attribute_group = working_attribute_group

    def __repr__(self):
        return f"{self.__class__.__name__}({self.state}, {self.attribute_group}, {self.working_attribute_group})"


class _EntryParser(_Scope):
    """
    Moves the complexity and state management involved in parsing
    a full entry out of :class:`TextSpectrumLibrary`, allowing it
    to be factored into a bunch of helper methods around a single
    piece of shared stated too granular for the main parser.
    """

    library: "TextSpectralLibrary"
    state: _SpectrumParserStateEnum

    spectrum: Optional[Spectrum]
    cluster: Optional[SpectrumCluster]
    analyte: Optional[Analyte]
    interpretation: Optional[Interpretation]
    interpretation_member: Optional[InterpretationMember]
    attribute_group: Optional[str]
    aggregation_types: List[str]
    peak_list: List[Tuple]

    start_line_number: int
    line_number: int = -1

    def __init__(
        self, library, start_line_number: int, spectrum_index: Optional[int]
    ) -> None:
        super().__init__(_SpectrumParserStateEnum.header, None)

        self.library = library
        self.start_line_number = start_line_number or 0
        self.spectrum_index = spectrum_index

        self.aggregation_types = None
        self.peak_list = []

        self.spectrum = None
        self.cluster = None
        self.analyte = None
        self.interpretation = None
        self.interpretation_member = None

    def _parse_attribute_into(
        self, line: str, store: Attributed, line_number_message: str
    ):
        self.library._parse_attribute_into(line, store, line_number_message, self)

    def real_line_number_or_nothing(self):
        message = f" on line {self.line_number + self.start_line_number}"
        if self.spectrum_index is not None:
            message += f" in spectrum index {self.spectrum_index}"
        message += f" in state {self.state}"
        return message

    def _parse_header(self, line):
        if START_OF_SPECTRUM_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.header
            self.spectrum = self.library._new_spectrum()
            self.spectrum.index = self.spectrum_index
            match = START_OF_SPECTRUM_MARKER.match(line)
            self.spectrum.key = int(match.group(1)) or self.spectrum.index - 1
            return

        elif START_OF_PEAKS_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.peaks
            return

        elif START_OF_INTERPRETATION_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.interpretation
            match = START_OF_INTERPRETATION_MARKER.match(line)
            if self.interpretation is not None:
                self.spectrum.add_interpretation(self.interpretation)
            self.interpretation = self.library._new_interpretation(match.group(1))
            self.spectrum.add_interpretation(self.interpretation)
            self.analyte = None
            return

        elif START_OF_ANALYTE_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.analyte
            match = START_OF_ANALYTE_MARKER.match(line)
            self.analyte = self.library._new_analyte(match.group(1))
            self.spectrum.add_analyte(self.analyte)
            return

        elif START_OF_CLUSTER.match(line):
            self.state = _SpectrumParserStateEnum.cluster
            self.cluster = self.library._new_cluster()
            match = START_OF_CLUSTER.match(line)
            self.cluster.key = int(match.group(1)) or self.cluster.index - 1
            return

        if self.spectrum is None:
            raise ValueError(
                f"Spectrum not yet initialized when trying to add attribute in state {self.state}{self.real_line_number_or_nothing()}"
            )
        self._parse_attribute_into(
            line, self.spectrum, self.real_line_number_or_nothing
        )

    def _parse_interpretation(self, line):
        if START_OF_ANALYTE_MARKER.match(line):
            warnings.warn(
                f"An analyte found after an interpretation was encountered, {self.real_line_number_or_nothing()}"
            )
            self.state = _SpectrumParserStateEnum.analyte
            match = START_OF_ANALYTE_MARKER.match(line)
            if self.analyte is not None:
                self.spectrum.add_analyte(self.analyte)
            self.analyte = self.library._new_analyte(match.group(1))
            self.spectrum.add_analyte(self.analyte)
            return
        elif START_OF_INTERPRETATION_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.interpretation
            match = START_OF_INTERPRETATION_MARKER.match(line)
            if self.interpretation is not None:
                self.spectrum.add_interpretation(self.interpretation)
            self.interpretation = self.library._new_interpretation(match.group(1))
            self.spectrum.add_interpretation(self.interpretation)
            self.analyte = None
            return
        elif START_OF_PEAKS_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.peaks
            return
        elif START_OF_INTERPRETATION_MEMBER_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.interpretation_member
            match = START_OF_INTERPRETATION_MEMBER_MARKER.match(line)

            if self.interpretation_member is not None:
                self.interpretation.add_member_interpretation(
                    self.interpretation_member
                )

            self.interpretation_member = InterpretationMember(match.group(1))
            self.interpretation.add_member_interpretation(self.interpretation_member)
            return

        self._parse_attribute_into(
            line, self.interpretation.attributes, self.real_line_number_or_nothing
        )
        self.library._analyte_interpretation_link(self.spectrum, self.interpretation)

    def _parse_interpretation_member(self, line):
        if START_OF_PEAKS_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.peaks
            self.interpretation_member = None
            self.interpretation = None
            return
        elif START_OF_INTERPRETATION_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.interpretation
            match = START_OF_INTERPRETATION_MARKER.match(line)
            if self.interpretation is not None:
                self.spectrum.add_interpretation(self.interpretation)
            self.interpretation = self.library._new_interpretation(match.group(1))
            self.spectrum.add_interpretation(self.interpretation)
            self.interpretation_member = None
            return
        elif START_OF_INTERPRETATION_MEMBER_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.interpretation_member
            match = START_OF_INTERPRETATION_MEMBER_MARKER.match(line)
            if self.interpretation_member is not None:
                self.interpretation.add_member_interpretation(
                    self.interpretation_member
                )
            self.interpretation_member = InterpretationMember(match.group(1))
            self.interpretation.add_member_interpretation(self.interpretation_member)
            return

        self._parse_attribute_into(
            line, self.interpretation_member, self.real_line_number_or_nothing
        )

    def _parse_analyte(self, line):
        if START_OF_PEAKS_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.peaks
            if self.analyte is not None:
                self.spectrum.add_analyte(self.analyte)
                self.analyte = None
            return

        elif START_OF_ANALYTE_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.analyte
            match = START_OF_ANALYTE_MARKER.match(line)
            if self.analyte is not None:
                self.spectrum.add_analyte(self.analyte)
            self.analyte = self.library._new_analyte(match.group(1))
            return

        elif START_OF_INTERPRETATION_MARKER.match(line):
            self.state = _SpectrumParserStateEnum.interpretation
            match = START_OF_INTERPRETATION_MARKER.match(line)
            if self.analyte is not None:
                self.spectrum.add_analyte(self.analyte)
                self.analyte = None

            # Somehow we have an in-progress Interpretation that hasn't been cleared yet.
            # This should probably be an error strictly speaking.
            if self.interpretation is not None:
                warnings.warn(
                    f"Interleaved analytes and interpretations detected at {self.real_line_number_or_nothing()}"
                )
                self.spectrum.add_interpretation(self.interpretation)
            self.interpretation = self.library._new_interpretation(match.group(1))
            self.spectrum.add_interpretation(self.interpretation)
            return

        self._parse_attribute_into(line, self.analyte, self.real_line_number_or_nothing)

    def _parse_peaks(self, line):
        # TODO: When we know more about how different aggregations are formatted,
        # look that up here once so we remember it and can use it to process the
        # aggregation columns
        if self.aggregation_types is None:
            self.aggregation_types = self.spectrum.peak_aggregations
        match = float_number.match(line)
        if match is not None:
            tokens = line.split("\t")
            n_tokens = len(tokens)
            if n_tokens == 1 and " " in line:
                if match := FALLBACK_PEAK_LINE_PATTERN.match(line):
                    tokens = match.groups()
                    n_tokens = len(tokens)
                    warnings.warn(
                        f"Space character delimiter found in peak line{self.real_line_number_or_nothing()}",
                        ValidationWarning,
                    )
            if n_tokens == 2:
                mz, intensity = tokens
                annotation = parse_annotation("?")
                self.peak_list.append([float(mz), float(intensity), annotation, []])
            elif n_tokens == 3:
                mz, intensity, annotation = tokens
                if not annotation:
                    annotation = "?"
                annotation = parse_annotation(annotation, wrap_errors=True)
                self.peak_list.append([float(mz), float(intensity), annotation, []])
            elif n_tokens > 3:
                mz, intensity, annotation, *aggregation = tokens
                if not annotation:
                    annotation = "?"
                annotation = parse_annotation(annotation, wrap_errors=True)
                self.peak_list.append(
                    [
                        float(mz),
                        float(intensity),
                        annotation,
                        [try_cast(agg) for agg in aggregation],
                    ]
                )
            else:
                raise ValueError(
                    f"Malformed peak line {line} with {n_tokens} entries{self.real_line_number_or_nothing()}"
                )
        else:
            raise ValueError(
                f"Malformed peak line {line}{self.real_line_number_or_nothing()}"
            )

    def _parse_cluster(self, line):
        if START_OF_SPECTRUM_MARKER.match(line):
            raise ValueError(
                f"Clusters should not include spectrum sections {self.real_line_number_or_nothing()}"
            )

        elif START_OF_PEAKS_MARKER.match(line):
            raise ValueError(
                f"Clusters should not include peaks {self.real_line_number_or_nothing()}"
            )

        elif START_OF_INTERPRETATION_MARKER.match(line):
            raise ValueError(
                f"Clusters should not include interpretation sections {self.real_line_number_or_nothing()}"
            )

        elif START_OF_ANALYTE_MARKER.match(line):
            raise ValueError(
                f"Clusters should not include analyte sections {self.real_line_number_or_nothing()}"
            )

        elif START_OF_INTERPRETATION_MEMBER_MARKER.match(line):
            raise ValueError(
                f"Clusters should not include interpretation member sections {self.real_line_number_or_nothing()}"
            )

        self._parse_attribute_into(line, self.cluster, self.real_line_number_or_nothing)

    def parse(self, buffer: Iterable[str]):
        line: str
        for line_number, line in enumerate(buffer):
            self.line_number = line_number
            line = line.strip()
            if not line:
                break
            # Skip comments for now, no round-trip
            if line.startswith("#"):
                continue
            elif self.state == _SpectrumParserStateEnum.header:
                self._parse_header(line)
            elif self.state == _SpectrumParserStateEnum.interpretation:
                self._parse_interpretation(line)
            elif self.state == _SpectrumParserStateEnum.interpretation_member:
                self._parse_interpretation_member(line)
            elif self.state == _SpectrumParserStateEnum.analyte:
                self._parse_analyte(line)
            elif self.state == _SpectrumParserStateEnum.peaks:
                self._parse_peaks(line)
            elif self.state == _SpectrumParserStateEnum.cluster:
                self._parse_cluster(line)
            else:
                raise ValueError(
                    f"Unknown state {self.state}{self.real_line_number_or_nothing()}"
                )
        if self.cluster:
            return self.cluster
        self.spectrum.peak_list = self.peak_list
        # Backfill analytes into interpretations that never explicitly listed them.
        self.library._default_interpretation_to_analytes(self.spectrum)
        return self.spectrum


def _is_header_line(line: Union[str, bytes]) -> bool:
    if isinstance(line, bytes):
        line = line.decode("utf8")
    if START_OF_SPECTRUM_MARKER.match(line):
        return False
    if START_OF_CLUSTER.match(line):
        return False
    if SPECTRUM_NAME_PRESENT.match(line):
        return False
    return True


class TextSpectralLibrary(_PlainTextSpectralLibraryBackendBase):
    """
    A reader for the plain text serialization of the mzSpecLib spectral library foramt.

    This implementation may operate on a stream opened in binary mode or a file path.
    If using a non-seekable stream, the random access or search methods may not be
    supported.
    """

    file_format: ClassVar[List[str]] = ["mzlb.txt", "mzlib.txt"]
    format_name: ClassVar[str] = "text"

    @classmethod
    def guess_from_header(cls, filename: str) -> bool:
        with open_stream(filename, "r", encoding="utf8") as stream:
            first_line = stream.readline()
            if (
                START_OF_SPECTRUM_MARKER.match(first_line)
                or START_OF_LIBRARY_MARKER.match(first_line)
                or START_OF_CLUSTER.match(first_line)
            ):
                return True
        return False

    def _parse_header_from_stream(self, stream: io.RawIOBase) -> Tuple[bool, int]:
        nbytes = 0
        first_line = stream.readline()
        nbytes += len(first_line)
        first_line = first_line.decode("utf8")
        empty_checkpoint = None
        state = _LibraryParserStateEnum.unknown

        current_attribute_set = None
        current_attribute_set_type = None
        if not _is_header_line(first_line):
            return True, 0
        elif START_OF_LIBRARY_MARKER.match(first_line):
            state = _LibraryParserStateEnum.header
            match = START_OF_LIBRARY_MARKER.match(first_line)
            attributes = AttributeManager()
            version_in_header = match.group(1)
            if version_in_header:
                warnings.warn(
                    f"Library format header contains version tag {version_in_header}, this is no longer "
                    "part of the specification and is ignored.",
                    category=ValidationWarning,
                )
            line = stream.readline()
            while _is_header_line(line):
                nbytes += len(line)
                line = line.strip().decode("utf8")
                if not line:
                    if empty_checkpoint is None:
                        empty_checkpoint = stream.tell()
                    else:
                        if empty_checkpoint == stream.tell():
                            break
                        else:
                            empty_checkpoint = None
                    line = stream.readline()
                    continue
                match = START_OF_ATTRIBUTE_SET.match(line)
                if match:
                    state = _LibraryParserStateEnum.attribute_sets
                    if current_attribute_set is not None:
                        self._add_attribute_set(
                            current_attribute_set, current_attribute_set_type
                        )

                    current_attribute_set_type = attribute_set_types[
                        match.group(1).lower()
                    ]
                    attrib_set_name = match.group(2)
                    current_attribute_set = AttributeSet(attrib_set_name, [])
                else:
                    match = key_value_term_pattern.match(line)

                    # We found a single attribute
                    if match is not None:
                        d = match.groupdict()
                        self._prepare_attribute_dict(d)
                        # If we're in an attribute set, store it in the attribute set
                        if state == _LibraryParserStateEnum.attribute_sets:
                            current_attribute_set.add_attribute(d["term"], d["value"])
                        else:  # Otherwise store it in the library level attributes
                            if d["term"] == FORMAT_VERSION: # Don't add the format version attribute multiple times
                                if attributes.has_attribute(FORMAT_VERSION):
                                    logger.debug(
                                        "Encountered extra format version term with value %r, current value is %r, skipping",
                                        d["value"],
                                        attributes.get_attribute(FORMAT_VERSION),
                                    )
                                else:
                                    attributes.add_attribute(d["term"], d["value"])
                            else:
                                attributes.add_attribute(d["term"], d["value"])

                        line = stream.readline()
                        continue

                    if line.startswith("["):
                        # We found a grouped attribute
                        match = grouped_key_value_term_pattern.match(line)
                        if match is not None:
                            d = match.groupdict()
                            self._prepare_attribute_dict(d)
                            # If we're in an attribute set, store it in the attribute
                            # set
                            if state == _LibraryParserStateEnum.attribute_sets:
                                current_attribute_set.add_attribute(
                                    d["term"], d["value"], d["group_id"]
                                )
                                current_attribute_set.group_counter = int(d["group_id"])
                            else:  # Otherwise store it in the library level attributes
                                attributes.add_attribute(
                                    d["term"], d["value"], d["group_id"]
                                )
                                attributes.group_counter = int(d["group_id"])

                            line = stream.readline()
                            continue
                        else:
                            raise ValueError(f"Malformed grouped attribute {line}")
                    elif "=" in line:
                        name, value = line.split("=", 1)
                        if state == _LibraryParserStateEnum.attribute_sets:
                            current_attribute_set.add_attribute(name, value)
                        else:
                            attributes.add_attribute(name, value)
                    else:
                        raise ValueError(f"Malformed attribute line {line}")
                line = stream.readline()

            if current_attribute_set is not None:
                self._add_attribute_set(
                    current_attribute_set, current_attribute_set_type
                )
            if not attributes.has_attribute(FORMAT_VERSION):
                warnings.warn(f"Library does not have a {FORMAT_VERSION}, assuming current version", category=ValidationWarning)
                attributes = [Attribute(FORMAT_VERSION, DEFAULT_VERSION)] + list(attributes)
            self.attributes.clear()
            self.attributes._from_iterable(attributes)
            return True, nbytes
        return False, 0

    def read_header(self) -> bool:
        if isinstance(self.filename, io.IOBase) and not self.filename.seekable():
            return self._parse_header_from_stream(self.filename)[0]
        with open_stream(self.filename, "rb") as stream:
            return self._parse_header_from_stream(stream)[0]

    def _get_file_size(self):
        if isinstance(self.filename, (str, bytes, os.PathLike)):
            return os.path.getsize(self.filename)
        elif (
            hasattr(self.filename, "seek")
            and hasattr(self.filename, "tell")
            and self.filename.seekable()
        ):
            saved = self.filename.tell()
            self.filename.seek(0, os.SEEK_END)
            z = self.filename.tell()
            self.filename.seek(saved)
            return z
        else:
            return float("inf")

    def create_index(self) -> int:
        """
        Populate the spectrum index

        Returns
        -------
        n_spectra: int
            The number of entries read
        """
        #### Check that the spectrum library filename isvalid
        filename = self.filename

        #### Determine the filesize
        file_size = self._get_file_size()

        with open_stream(filename, "rt", encoding="utf8") as infile:
            state = "header"
            entry_buffer = deque()

            n_spectra = 0
            n_clusters = 0

            start_index = 0
            file_offset = 0

            line_beginning_file_offset = 0
            spectrum_file_offset = 0
            spectrum_name = ""
            current_key = None
            entry_is_cluster = False

            # Required for counting file_offset manually (LF vs CRLF)
            # NOTE: If the file has been manually editted with different line endings
            # for some but not all spectra, this assumption **WILL** probably fail
            # to maintain the invariant. We probably need something more intrusive.
            infile.readline()
            file_offset_line_ending = len(infile.newlines) - 1
            infile.seek(0)

            logger.debug(f"Reading {filename} ({file_size} bytes)...")
            while 1:
                line = infile.readline()
                if len(line) == 0:
                    break

                line_beginning_file_offset = file_offset

                # From Eric: tell() is twice as slow as counting it myself
                file_offset += len(line) + file_offset_line_ending

                line = line.rstrip()
                if state == "header":

                    if is_spec := START_OF_SPECTRUM_MARKER.match(line):
                        current_key = int(is_spec.group(1))
                        state = "body"
                        spectrum_file_offset = line_beginning_file_offset
                        entry_is_cluster = False
                    elif is_clus := START_OF_CLUSTER.match(line):
                        current_key = int(is_clus.group(1))
                        state = "body"
                        spectrum_file_offset = line_beginning_file_offset
                        entry_is_cluster = True
                    else:
                        continue

                if state == "body":
                    if len(line) == 0:
                        continue

                    is_spec = START_OF_SPECTRUM_MARKER.match(line)
                    is_clus = START_OF_CLUSTER.match(line)
                    if (is_spec) or (is_clus):
                        if len(entry_buffer) > 0:
                            if not entry_is_cluster:
                                if not spectrum_name:
                                    raise ValueError("No spectrum name")
                                self.index.add(
                                    number=current_key,
                                    offset=spectrum_file_offset,
                                    name=spectrum_name,
                                    analyte=None,
                                )
                                n_spectra += 1
                                current_key = (
                                    int(is_spec.group(1))
                                    if is_spec
                                    else int(is_clus.group(1))
                                )
                                #### Commit every now and then
                                if n_spectra % 10000 == 0:
                                    self.index.commit()
                                    logger.info(
                                        f"Processed {file_offset} bytes, {n_spectra} spectra read, {n_clusters} clusters read"
                                    )
                            else:
                                self.index.add_cluster(
                                    number=current_key, offset=spectrum_file_offset
                                )
                                if n_clusters % 10000 == 0:
                                    self.index.commit()
                                    logger.info(
                                        f"Processed {file_offset} bytes, {n_spectra} spectra read, {n_clusters} clusters read"
                                    )
                                n_clusters += 1
                                current_key = (
                                    int(is_spec.group(1))
                                    if is_spec
                                    else int(is_clus.group(1))
                                )

                        entry_buffer.clear()
                        entry_is_cluster = bool(is_clus)
                        spectrum_file_offset = line_beginning_file_offset
                        spectrum_name = ""

                    if SPECTRUM_NAME_PRESENT.match(line):
                        if match := SPECTRUM_NAME_MATCH.match(line):
                            spectrum_name = match.group(1)

                    entry_buffer.append(line)

            if spectrum_name:
                self.index.add(
                    number=current_key,
                    offset=spectrum_file_offset,
                    name=spectrum_name,
                    analyte=None,
                )
                self.index.commit()
                n_spectra += 1
                logger.info(
                    f"Processed {file_offset} bytes, {n_spectra} spectra read, {n_clusters} clusters read"
                )
            elif entry_is_cluster:
                self.index.add_cluster(
                    number=current_key,
                    offset=spectrum_file_offset,
                )
                self.index.commit()
                n_clusters += 1
                logger.info(
                    f"Processed {file_offset} bytes, {n_spectra} spectra read, {n_clusters} clusters read"
                )

            #### Flush the index
            self.index.commit()

        return n_spectra

    def _buffer_from_stream(self, infile: Iterable[str]) -> List:
        state = "body"
        spectrum_buffer = []

        for line in infile:
            line = line.rstrip()
            if state == "body":
                if len(line) == 0:
                    continue
                if START_OF_SPECTRUM_MARKER.match(line) or START_OF_CLUSTER.match(line):
                    if len(spectrum_buffer) > 0:
                        return spectrum_buffer
                spectrum_buffer.append(line)
        return spectrum_buffer

    def _prepare_attribute_dict(self, match):
        key = match["term_accession"]
        value = match["value"]
        try:
            term = self.find_term_for(key)
            match["value"] = term.value_type(value)
        except KeyError:
            match["value"] = try_cast(value)

    def _map_attribute_set(self, name: str, scope: _Scope) -> AttributeSet:
        attr_set = None
        if _SpectrumParserStateEnum.header == scope.state:
            attr_set = self.spectrum_attribute_sets[name]
        elif _SpectrumParserStateEnum.analyte == scope.state:
            attr_set = self.analyte_attribute_sets[name]
        elif _SpectrumParserStateEnum.interpretation == scope.state:
            attr_set = self.interpretation_attribute_sets[name]
        elif _SpectrumParserStateEnum.cluster == scope.state:
            attr_set = self.cluster_attribute_sets[name]
        else:
            raise ValueError(f"Cannot define attribute sets for {scope.state}")
        return attr_set

    def _parse_attribute(
        self, line: str, line_number_message=lambda: "", scope: Optional[_Scope] = None
    ) -> Union[Attribute, AttributeSet]:
        match = key_value_term_pattern.match(line)
        if scope is None:
            scope = _Scope(None, None)
        if match is not None:
            d = match.groupdict()
            self._prepare_attribute_dict(d)
            if d["term"] == ATTRIBUTE_SET_NAME:
                attr_set = self._map_attribute_set(d['value'], scope)
                return AttributeSetRef(attr_set, None)
            attr = Attribute(d["term"], d["value"])
            return attr
        elif line.startswith("["):
            match = grouped_key_value_term_pattern.match(line)
            if match is not None:
                d = match.groupdict()
                self._prepare_attribute_dict(d)
                if d["term"] == ATTRIBUTE_SET_NAME:
                    attr_set = self._map_attribute_set(d["value"], scope)
                    return AttributeSetRef(attr_set, d['group_id'])
                attr = Attribute(d["term"], d["value"], d["group_id"])
                return attr
            else:
                raise ValueError(
                    f"Malformed grouped attribute {line}{line_number_message()}"
                )
        elif "=" in line:
            name, value = line.split("=", 1)
            attr = Attribute(name, try_cast(value))
            return attr
        else:
            raise ValueError(f"Malformed attribute line {line}{line_number_message()}")

    def _parse_attribute_into(
        self,
        line: str,
        store: Attributed,
        line_number_message=lambda: "",
        scope: Optional[_Scope] = None,
    ) -> bool:
        if scope is None:
            scope = _Scope(None, None)
        attr = self._parse_attribute(line, line_number_message, scope)
        if isinstance(attr, (AttributeSet, AttributeSetRef)):
            group_id = getattr(attr, 'group_id', None)
            if group_id is not None:
                if group_id != scope.attribute_group:
                    scope.attribute_group = group_id
                    scope.working_attribute_group = store.get_next_group_identifier()
                    attr.group_id = scope.working_attribute_group
                else:
                    attr.group_id = scope.working_attribute_group
            store.add_attribute(
                ATTRIBUTE_SET_NAME,
                attr.name,
                group_identifier=attr.group_id
            )
            attr.apply(store, attr.group_id)
        else:
            if attr.group_id:
                if attr.group_id != scope.attribute_group:
                    scope.attribute_group = attr.group_id
                    scope.working_attribute_group = store.get_next_group_identifier()
                    attr.group_id = scope.working_attribute_group
                else:
                    attr.group_id = scope.working_attribute_group
            store.add_attribute(attr.key, attr.value, attr.group_id)
        return True

    def _parse(
        self,
        buffer: Iterable[str],
        spectrum_index: int = None,
        start_line_number: int = None,
    ) -> Union[Spectrum, SpectrumCluster]:
        parser = _EntryParser(self, start_line_number, spectrum_index)
        return parser.parse(buffer)

    def get_spectrum(
        self, spectrum_number: Optional[int] = None, spectrum_name: Optional[str] = None
    ) -> Spectrum:
        # keep the two branches separate for the possibility that this is not
        # possible with all index schemes.
        if spectrum_number is not None:
            if spectrum_name is not None:
                raise ValueError("Provide only one of spectrum_number or spectrum_name")
            index_record = self.index.record_for(spectrum_number)
            offset = index_record.offset
        elif spectrum_name is not None:
            index_record = self.index.record_for(spectrum_name)
            offset = index_record.offset
            spectrum_number = index_record.number
        else:
            raise ValueError(
                "Must provide either spectrum_number or spectrum_name argument"
            )

        buffer = self._get_lines_for(offset)
        spectrum = self._parse(buffer, index_record.index)
        return spectrum

    def get_cluster(self, cluster_number: int) -> SpectrumCluster:
        offset = self.index.offset_for_cluster(cluster_number)
        buffer = self._get_lines_for(offset)
        cluster = self._parse(buffer, cluster_number)
        return cluster


class TextSpectralLibraryWriter(SpectralLibraryWriterBase):
    """
    Write a spectral library to the plain text serialization of the mzSpecLib spectral library format.

    Attributes
    ----------
    version : str
        The format version to write in semver-compatible notation
    compact_interpretation : bool, default :const:`True`
        Whether to elect to write compact interpretation member sections when there is only
        one interpretation and only one interpretation member by inlining the interpretation
        member attributes into the interpretation. Both forms are valid, one is just less
        verbose.
    """

    file_format = "mzlb.txt"
    format_name = "text"
    default_version = "1.0"

    def __init__(
        self, filename, version=None, compact_interpretations: bool = True, **kwargs
    ):
        super(TextSpectralLibraryWriter, self).__init__(filename)
        self.version = version
        self._coerce_handle(self.filename)
        self.compact_interpretations = compact_interpretations

    def _write_attributes(self, attributes: Attributed, attribute_sets: Optional[List[str]]=None):
        if attribute_sets is None:
            attribute_sets = []
        for attribute in attributes:
            if attribute.owner_id and attribute.owner_id in attribute_sets or attribute.owner_id == "all":
                continue
            value = attribute.value
            if ":" in attribute.key:
                try:
                    term = self.find_term_for(attribute.key.split("|")[0])
                    value = term.value_type.format(value)
                except (KeyError, ValueError):
                    pass
            if attribute.group_id is None:
                self.handle.write(f"{attribute.key}={value}\n")
            else:
                self.handle.write(f"[{attribute.group_id}]{attribute.key}={value}\n")

    def write_header(self, library: SpectralLibraryBackendBase):
        """Write the library header and other global metadata"""
        self.handle.write("<mzSpecLib>\n")
        self._write_attributes(library.attributes)
        for attr_set in library.spectrum_attribute_sets.values():
            self._write_attribute_set(attr_set, AttributeSetTypes.spectrum)

        for attr_set in library.analyte_attribute_sets.values():
            self._write_attribute_set(attr_set, AttributeSetTypes.analyte)

        for attr_set in library.interpretation_attribute_sets.values():
            self._write_attribute_set(attr_set, AttributeSetTypes.interpretation)

    def _write_attribute_set(
        self, attribute_set: AttributeSet, attribute_set_type: AttributeSetTypes
    ):
        if attribute_set_type == AttributeSetTypes.spectrum:
            set_type = "Spectrum"
        elif attribute_set_type == AttributeSetTypes.analyte:
            set_type = "Analyte"
        elif attribute_set_type == AttributeSetTypes.interpretation:
            set_type = "Interpretation"
        elif attribute_set_type == AttributeSetTypes.cluster:
            set_type = "Cluster"

        header = f"<AttributeSet {set_type}={attribute_set.name}>\n"
        self.handle.write(header)
        self._write_attributes(attribute_set.attributes)
        if attribute_set.attributes:
            self.handle.write("\n")

    def write_spectrum(self, spectrum: Spectrum):
        self.handle.write(f"<Spectrum={spectrum.key}>\n")
        attribs_of = list(
            self._filter_attributes(spectrum, self._not_entry_key_or_index)
        )
        self._write_attributes(attribs_of, spectrum.attribute_sets)
        for analyte in spectrum.analytes.values():
            self.handle.write(f"<Analyte={analyte.id}>\n")
            self._write_attributes(analyte.attributes, analyte.attribute_sets)
        _n_interps = len(spectrum.interpretations)
        for interpretation in spectrum.interpretations.values():
            interpretation: Interpretation

            if len(spectrum.analytes) == 1:
                attribs_of = list(
                    self._filter_attributes(
                        interpretation, self._not_analyte_mixture_term
                    )
                )
            else:
                attribs_of = interpretation.attributes

            if len(attribs_of) == 0 and len(spectrum.interpretations) == 1:
                continue

            self.handle.write(f"<Interpretation={interpretation.id}>\n")
            self._write_attributes(attribs_of, interpretation.attribute_sets)

            # When there is only one interpretation and only one interpretation member
            # interpretation member attributes are written out as part of the interpretation
            # itself.
            if (
                _n_interps == 1
                and len(interpretation.member_interpretations) == 1
                and self.compact_interpretations
            ):
                for member in interpretation.member_interpretations.values():
                    self._write_attributes(member.attributes, member.attribute_sets)
            else:
                for member in interpretation.member_interpretations.values():
                    member: InterpretationMember
                    self.handle.write(f"<InterpretationMember={member.id}>\n")
                    self._write_attributes(member.attributes, member.attribute_sets)
        self.handle.write("<Peaks>\n")
        for peak in spectrum.peak_list:
            peak_parts = [
                str(peak[0]),
                str(peak[1]),
                "?" if not peak[2] else ",".join(map(str, peak[2])),
            ]
            if peak[3]:
                peak_parts.append("\t".join(map(_format_aggregation, peak[3])))
            self.handle.write("\t".join(peak_parts) + "\n")
        self.handle.write("\n")

    def write_cluster(self, cluster: SpectrumCluster):
        self.handle.write(f"<Cluster={cluster.key}>\n")
        attribs_of = list(
            self._filter_attributes(cluster, self._not_entry_key_or_index)
        )
        self._write_attributes(attribs_of)
        self.handle.write("\n")

    def close(self):
        if not self.handle.closed:
            self.handle.close()


def _format_aggregation(value: Union[numbers.Number, str]) -> str:
    if isinstance(value, numbers.Number):
        return "%0.4g" % value
    else:
        return value


def format_spectrum(spectrum: Spectrum, **kwargs) -> str:
    buffer = io.StringIO()
    writer = TextSpectralLibraryWriter(buffer, **kwargs)
    writer.write_spectrum(spectrum)
    return buffer.getvalue()



import os

data_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "test_data"))


def datafile(name):
    path = os.path.join(data_path, name)
    if not os.path.exists(path):
        raise Exception(
            ("The path %r could not be located in the test suite's data package." % (path, )) +
            "If you are NOT running the test suite, you should not"
            " be using this function.")
    return path



import dataclasses
import io
import json
import logging
import re

from datetime import datetime
from importlib import resources

from xml.etree import ElementTree as etree

from typing import Any, ClassVar, Dict, List, TYPE_CHECKING, Mapping, Optional, Sequence, Set, Tuple, Union

from mzspeclib.attributes import Attributed
from mzspeclib.utils import flatten, ensure_iter
from mzspeclib.ontology import _VocabularyResolverMixin

from .level import RequirementLevel, CombinationLogic

if TYPE_CHECKING:
    from .validator import ValidatorBase


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class AttributeSemanticPredicate:
    """
    A predicate rule that applies to a :class:`~.Attribute` value.

    This is a base class with no specific validation behavior of its
    own. See its children for specific kinds of predicates.
    """

    name: ClassVar[str]

    _registry = {}

    def validate(self, attribute: 'AttributeSemanticRule', value: str, validator_context: "ValidatorBase") -> bool:
        raise NotImplementedError()

    def to_dict(self) -> Dict[str, Any]:
        raise NotImplementedError()

    @classmethod
    def from_dict(cls, state: Dict[str, Any]) -> 'AttributeSemanticPredicate':
        if isinstance(state, Mapping):
            name = state['name']
        elif isinstance(state, str):
            name = state
        else:
            raise TypeError(f"Cannot convert {state} to {cls.__name__}")
        rule_tp = cls._registry[name]
        return rule_tp.from_dict(state)

    def __init_subclass__(cls, *args, **kwargs):
        cls._registry[cls.name] = cls
        super().__init_subclass__(**kwargs)


class ValueOfType(AttributeSemanticPredicate):
    """Require that an attribute's value's type is one of the permited kinds."""

    type_name: Union[str, List[str]]

    name = "value_of_type"

    def __init__(self, type_name):
        super().__init__()
        self.type_name = type_name

    def validate(self, attribute: 'AttributeSemanticRule', value: str, validator_context: "ValidatorBase") -> bool:
        if isinstance(value, list) and attribute.repeatable:
            if all(self.validate(attribute, v, validator_context) for v in value):
                return True
        if not isinstance(self.type_name, list):
            type_name = [self.type_name]
        else:
            type_name = self.type_name
        result = False
        for tp in type_name:
            if tp == 'int':
                result |= isinstance(value, int) or (isinstance(value, float) and value.is_integer())
            elif tp == 'float':
                result |= isinstance(value, float)
            elif tp == "string":
                result |= isinstance(value, str)
            elif tp == 'datetime':
                if isinstance(value, datetime.datetime):
                    result |= True
                elif isinstance(value, str):
                    logger.warning("Ambiguous datetime found for %s: %s", attribute, value)
                    result |= True
            else:
                raise ValueError(f"Can't validate type {tp}")
        return result

    def to_dict(self) -> Dict[str, Any]:
        state = {
            "name": self.name,
            "type_name": self.type_name
        }
        return state

    @classmethod
    def from_dict(cls, state: Dict[str, Any]) -> 'AttributeSemanticPredicate':
        return cls(state['type_name'])


class ValueIsChildOf(AttributeSemanticPredicate):
    """
    Require that the attribute's value is semantically derived from the controlled
    vocabulary term given by :attr:`accession`.
    """

    accession: str

    name = "value_is_child_of"

    def __init__(self, accession):
        super().__init__()
        self.accession = accession

    def validate(self, attribute: 'AttributeSemanticRule', value: str, validator_context: "ValidatorBase"):
        if isinstance(value, list) and attribute.repeatable:
            return all(self.validate(attribute, v, validator_context) for v in value)
        for term in validator_context.walk_terms_for(self.accession):
            if term == value:
                if not term.startswith(self.accession):
                    return True
        return False

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "accession": self.accession
        }

    @classmethod
    def from_dict(cls, state: Dict[str, Any]) -> 'AttributeSemanticPredicate':
        return cls(state['accession'])


class ValueIsUnique(AttributeSemanticPredicate):
    """
    Ensure that the value of this attribute is **unique** throughout the library.

    .. note::

        This will store all values encountered in memory, which may be costly
        for large libraries.

    """

    seen: set

    name = "value_is_unique"

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.seen = set()

    def validate(self, attribute: 'AttributeSemanticRule', value: str, validator_context: "ValidatorBase"):
        if isinstance(value, list):
            if attribute.repeatable:
                return all(self.validate(attribute, v, validator_context) for v in value)
            else:
                return False
        if value in self.seen:
            return False
        self.seen.add(value)
        return True

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
        }

    @classmethod
    def from_dict(cls, state: Dict[str, Any]) -> 'AttributeSemanticPredicate':
        return cls()


class ValueMatches(AttributeSemanticPredicate):
    """Ensure that the value of an attribute matches a specific term"""

    accession: str

    name = "value_matches"

    def __init__(self, accession):
        super().__init__()
        self.accession = accession

    def validate(self, attribute: 'AttributeSemanticRule', value: str, validator_context: "ValidatorBase"):
        if isinstance(value, list) and attribute.repeatable:
            return all(self.validate(attribute, v, validator_context) for v in value)
        entity = validator_context.find_term_for(self.accession)
        key = f"{entity.id}|{entity.name}"
        return key == value

    def to_dict(self) -> Dict[str, Any]:
        state = {
            "name": self.name,
            "accession": self.accession
        }
        return state

    @classmethod
    def from_dict(cls, state: Dict[str, Any]) -> 'AttributeSemanticPredicate':
        return cls(state['accession'])


class ValueMatchesPattern(AttributeSemanticPredicate):
    """Ensure that an attribute's value matches a specified regular expression"""

    pattern: re.Pattern

    name = "value_matches_pattern"

    def __init__(self, pattern: str):
        super().__init__()
        self.pattern = re.compile(pattern)

    def to_dict(self) -> Dict[str, Any]:
        state = {}
        state['name'] = self.name
        state['pattern'] = self.pattern.pattern
        return state

    @classmethod
    def from_dict(cls, state: Dict[str, Any]) -> 'AttributeSemanticPredicate':
        return cls(state['pattern'])


@dataclasses.dataclass(frozen=True)
class AttributeSemanticRule:
    """
    A semantic validation rule that enforces constraints or
    requirements on a specific controlled vocabulary term-defined
    :class:`~.Attribute` in an entity.

    This may include enforcing a :class:`AttributeSemanticPredicate`
    on the value of the attribute, controlling how often an attribute
    may be used, or what units are assumed for it.
    """

    accession: str
    name: str
    repeatable: bool
    allow_children: bool
    value: Optional[AttributeSemanticPredicate] = dataclasses.field(default=None)
    condition: Optional['AttributeSemanticRule'] = dataclasses.field(default=None)
    notes: Optional[str] = dataclasses.field(default=None)
    default_unit: Optional[str] = dataclasses.field(default=None)

    @property
    def attribute(self) -> str:
        return f"{self.accession}|{self.name}"

    def to_dict(self) -> Dict[str, Any]:
        state = {
            "accession": self.accession,
            "name": self.name,
            "repeatable": self.repeatable,
            "allow_children": self.allow_children,
        }
        if self.value is not None:
            state['value'] = self.value.to_dict()
        if self.condition is not None:
            state['condition'] = self.condition.to_dict()
        if self.notes:
            state['notes'] = self.notes
        if self.default_unit is not None:
            state['default_unit'] = self.default_unit
        return state

    @classmethod
    def from_dict(cls, state, cv_provider: _VocabularyResolverMixin) -> 'AttributeSemanticRule':
        if isinstance(state, str):
            state = {
                "accession": state,
            }

        if "name" not in state:
            term = cv_provider.find_term_for(state)
            state['name'] = term.name

        repeatable = state.get("repeatable", False)
        allow_children = state.get("allow_children", False)

        value_rule = state.get("value")
        if value_rule:
            value_rule = AttributeSemanticPredicate.from_dict(value_rule)

        condition = state.get("condition")
        if condition:
            condition = cls.from_dict(condition, cv_provider)

        attr_rule = cls(
            state["accession"],
            state["name"],
            repeatable=repeatable,
            allow_children=allow_children,
            value=value_rule,
            condition=condition,
            notes=state.get('notes'),
            default_unit=state.get('default_unit'),
        )
        return attr_rule


@dataclasses.dataclass
class ScopedSemanticRule:
    """
    A semantic attribute rule that applies to a specific scope or context
    in a spectral library.

    Attributes
    ----------
    id : str
        A unique identifier for this rule
    path : str
        The validation path this rule applies to, e.g. /Library or /Library/Spectrum/Analyte
    requirement_level : :class:`~.RequirementLevel`
        How strong the requirement this rule be obeyed is
    combination_logic : :class:`~.CombinationLogic`
        How this rule's attribute rules interact
    attributes : list[:class:`AttributeSemanticRule`]
        The attribute rules applied by this semantic rule
    condition : :class:`AttributeSemanticRule`, optional
        A pre-condition rule that must be met for this rule to be applied
    notes : :class:`str`, optional
        Human-readable description of what this rule enforces or clarifies its intent
    """

    id: str
    path: str
    attributes: List[AttributeSemanticRule]
    requirement_level: RequirementLevel
    combination_logic: CombinationLogic
    condition: Optional[AttributeSemanticRule] = dataclasses.field(default=None)
    notes: Optional[str] = dataclasses.field(default=None)

    def __hash__(self):
        return hash(self.id)

    def find_all_children_of(self, attribute_rule: AttributeSemanticRule, obj: Attributed, validator_context: "ValidatorBase") -> Tuple:
        result = []
        for attrib in validator_context.walk_terms_for(attribute_rule.accession):
            try:
                hits = obj.get_attribute(attrib)
                if hits is not None:
                    result.append((attrib, hits))
            except KeyError:
                continue
        return tuple(result) if result else None

    def check_rule(self, obj: Attributed, attrib: AttributeSemanticRule, validator_context: "ValidatorBase") -> bool:
        result = True
        if attrib.allow_children:
            value = self.find_all_children_of(attrib, obj, validator_context)
        else:
            try:
                value = obj.get_attribute(attrib.attribute)
            except KeyError:
                value = None

        if isinstance(value, list) and not attrib.repeatable:
            result = False
        if attrib.value:
            if not attrib.value.validate(attrib, value, validator_context):
                result = False
        return result

    def validate(self, obj: Attributed, path: str, identifier_path: Tuple, validator_context: "ValidatorBase") -> bool:
        values = []

        if self.condition:
            if not self.check_rule(obj, self.condition, validator_context):
                return True

        for attrib in self.attributes:
            if attrib.allow_children:
                values.append(self.find_all_children_of(attrib, obj, validator_context))
            else:
                try:
                    values.append(obj.get_attribute(attrib.attribute))
                except KeyError:
                    values.append(None)

        result = True
        for attrib, value in zip(self.attributes, values):
            # When `get_attribute` matches an attribute multiple times, this will be
            # a list. `find_all_children_of` explicitly returns a tuple to avoid triggering
            # this.
            if isinstance(value, list) and not attrib.repeatable:
                validator_context.add_warning(
                    obj, path, self, identifier_path, value,
                    self.requirement_level,
                    f"{attrib.attribute} cannot be repeated")

                result = False
            if attrib.value:
                if not attrib.value.validate(attrib, value, validator_context):
                    validator_context.add_warning(
                        obj, path, identifier_path, self, value, self.requirement_level,
                        f"{self.id} required the value of {attrib.accession}|{attrib.name} conform to {attrib.value.name}")
                    result = False

        has_value = [v is not None for v in values]
        n_had_values = sum(has_value)

        if self.combination_logic == CombinationLogic.or_:
            if n_had_values < 1:
                validator_context.add_warning(
                    obj, path, identifier_path, self, value, self.requirement_level,
                    f"{self.id} requires one of {', '.join(attr.attribute for attr in self.attributes)}, none found")
                result = False

        elif self.combination_logic == CombinationLogic.and_:
            if n_had_values != len(self.attributes):
                missing = []
                for attr, v in zip(self.attributes, has_value):
                    if not v:
                        missing.append(attr)
                validator_context.add_warning(
                    obj, path, identifier_path, self, value, self.requirement_level,
                    f"{self.id} requires one of {', '.join(attr.attribute for attr in self.attributes)}, "
                    f"{', '.join(m.attribute for m in missing)} were absent")
                result = False

        elif self.combination_logic == CombinationLogic.xor:
            if n_had_values != 1:
                validator_context.add_warning(
                    obj, path, identifier_path, self, value, self.requirement_level,
                    f"{self.id} requires exactly one of {', '.join(attr.attribute for attr in self.attributes)}, "
                    "multiple were found.")
                result = False
        return result

    def __call__(self, obj: Attributed, path: str, identifier_path: Tuple, validator_context: "ValidatorBase") -> bool:
        return self.validate(obj, path, identifier_path, validator_context)

    @classmethod
    def from_xml(cls, stream: io.IOBase) -> List['ScopedSemanticRule']:
        tree = etree.parse(stream)
        rules = []
        for node in tree.findall(".//CvMappingRule"):
            term_rules = []
            for term in node.findall("CvTerm"):
                term_rules.append(
                    AttributeSemanticRule(
                        term.attrib['termAccession'],
                        term.attrib['termName'],
                        repeatable=term.attrib['isRepeatable'] == 'true',
                        allow_children=term.attrib['allowChildren'] == 'true'
                    )
                )
            attrib = node.attrib
            rule = ScopedSemanticRule(
                id=attrib['id'],
                path=attrib['scopePath'],
                requirement_level=RequirementLevel.from_str(attrib['requirementLevel']),
                combination_logic=CombinationLogic.from_str(attrib['cvTermsCombinationLogic']),
                attributes=term_rules
            )
            rules.append(rule)
        return rules

    @classmethod
    def from_dict(cls, data: Dict[str, Any], cv_provider: _VocabularyResolverMixin) -> List['ScopedSemanticRule']:
        rules = []
        for rule_spec in data['rules']:
            rule_id = rule_spec['id']
            level = RequirementLevel.from_str(rule_spec['level'].upper())
            path = rule_spec['path']
            combinator = CombinationLogic.from_str(
                rule_spec.get("combination_logic", "OR").upper())

            attribute_rules = []
            for attrib in rule_spec['attr']:
                attr_rule = AttributeSemanticRule.from_dict(attrib, cv_provider)
                attribute_rules.append(attr_rule)

            condition = rule_spec.get("condition")
            if condition:
                condition = AttributeSemanticRule.from_dict(condition, cv_provider)

            rules.append(
                ScopedSemanticRule(
                    id=rule_id,
                    path=path,
                    attributes=attribute_rules,
                    requirement_level=level,
                    combination_logic=combinator,
                    condition=condition,
                    notes=rule_spec.get('rules')
                )
            )
        return rules

    def to_dict(self) -> Dict[str, Any]:
        state = {
            "id": self.id,
            "path": self.path,
            "level": self.requirement_level.name.upper(),
            "combination_logic": self.combination_logic.to_str(),
            "attr": [
                a.to_dict() for a in self.attributes
            ],
            "notes": self.notes
        }
        if self.condition:
            state['condition'] = self.condition
        return state


class RuleSet(Sequence[ScopedSemanticRule]):
    name: str
    rules: List[ScopedSemanticRule]

    def __init__(self, name: str, rules: List[ScopedSemanticRule]):
        super().__init__()
        self.name = name
        self.rules = rules

    def __iter__(self):
        return iter(self.rules)

    def __len__(self):
        return len(self.rules)

    def __getitem__(self, i):
        return self.rules[i]

    @classmethod
    def from_dict(cls, data: Dict[str, any]) -> 'RuleSet':
        name = data['name']
        rules = ScopedSemanticRule.from_dict(data)
        return cls(name, rules)

    def to_dict(self) -> Dict[str, Any]:
        state = {
            "name": self.name,
            "rule": [
                rule.to_dict() for rule in self.rules
            ]
        }
        return state


def load_rule_set(name: str) -> List[ScopedSemanticRule]:
    res = resources.open_text(
        __name__.replace(".semantic_rule", '') + '.rules',
        name + '.json'
    )

    with res:
        rules = ScopedSemanticRule.from_dict(
            json.load(res),
            _VocabularyResolverMixin()
        )

    return RuleSet(
        name,
        rules
    )



import os
import unittest
import tempfile

from mzspeclib.backends import TextSpectralLibrary
from mzspeclib.index import MemoryIndex, SQLIndex

from .common import datafile

class IndexBehaviorBase(object):
    test_file = datafile("chinese_hamster_hcd_selected_head.mzlb.txt")
    index_cls = None

    def make_dir(self):
        dirname = tempfile.mkdtemp("_test", "mzlb_")
        return dirname

    def _open(self):
        return TextSpectralLibrary(self.test_file)

    def _make_index(self, library):
        library.index = self.index_cls()
        library.create_index()
        return library.index

    def test_sequence_behavior(self):
        lib = self._open()
        index = self._make_index(lib)
        assert len(index) == 7
        record = index[3]
        assert record.number == 4
        assert record.name == "AAAAGSTSVKPIFSR/2_0_44eV"


class TestMemoryIndex(unittest.TestCase, IndexBehaviorBase):
    index_cls = MemoryIndex


class TestSQLIndex(unittest.TestCase, IndexBehaviorBase):
    index_cls = SQLIndex

    def _make_index(self, library):
        dirname = self.make_dir()
        index_root = os.path.join(dirname, os.path.basename(self.test_file))
        library.index = self.index_cls(index_root)
        library.create_index()
        return library.index



from typing import Iterable, Any, Union, TypeVar

T = TypeVar("T")

def ensure_iter(obj: Union[Any, T, Iterable[T]]) -> Iterable[Union[Any, T]]:
    if isinstance(obj, Iterable):
        if not isinstance(obj, (str, bytes)):
            return obj
    return (obj, )


def flatten(obj: Iterable[Union[Iterable[T], T]]) -> Iterable[T]:
    acc = []
    for val in obj:
        if isinstance(val, Iterable) and not isinstance(val, (str, bytes)):
            acc.extend(flatten(val))
        else:
            acc.append(val)
    return acc


class ValidationWarning(UserWarning):
    """
    Indicates that something was parsed that did not halt the parser but
    which violates the expectations of the parser.

    The parser will make a best-effort attempt to interpret the value
    correctly but when validating this will count as a violation.
    """


import enum

class RequirementLevel(enum.IntEnum):
    may = 0
    should = 1
    must = 2
    disallow = 3

    @classmethod
    def from_str(cls, text: str) -> 'RequirementLevel':
        text = str(text).lower()
        return cls[text]


class CombinationLogic(enum.Enum):
    or_ = enum.auto()
    and_ = enum.auto()
    xor = enum.auto()

    @classmethod
    def from_str(cls, text: str) -> 'CombinationLogic':
        text = str(text).lower()
        if text in ("or", "and"):
            text += '_'
        return cls[text]

    def to_str(self) -> str:
        if self == self.or_:
            return "OR"
        elif self == self.and_:
            return "AND"
        elif self == self.xor:
            return "XOR"
        else:
            raise ValueError(self)



import os
import unittest
import tempfile

from mzspeclib.backends import MSPSpectralLibrary, TextSpectralLibrary

from .common import datafile


class TestSpectrum(unittest.TestCase):

    def get_library(self):
        test_file = datafile("chinese_hamster_hcd_selected_head.mzlb.txt")
        return TextSpectralLibrary(test_file)

    def get_spectrum(self, index):
        library = self.get_library()
        return library.get_spectrum(index)

    def test_write(self):
        spectrum = self.get_spectrum(1)
        buffer = spectrum.write('text')
        lines = buffer.splitlines()
        n_lines = len(lines)
        assert n_lines == 131
        assert buffer.startswith("<Spectrum=1>\nMS:1003061|library spectrum name")

    def test_equality(self):
        spectrum = self.get_spectrum(1)
        spectrum2 = self.get_spectrum(1)
        assert spectrum == spectrum2




"""
Spectral Library Indexing
=========================

:title-reference:`MzSpecLib` recommends that each file format be indexed in
a way that is appropriate to the format and the application, and
recognizes that most formats will not have any kind of built-in
index.

:mod:`mzspeclib` provides data structures for both in-memory and on-disk
index storage.

The :class:`~.IndexBase` type is an abstract base class for writing
indices. It isn't necessary unless you're writing a new backend or
index format.

The :class:`~.MemoryIndex` type holds all offset information and metadata
in memory, which can make it fast to access but problematic if several
large libraries are open at once. The index is not saved, requiring a
potentially costly scan of the entire library file when opening a file.

The :class:`~.SQLIndex` type holds its information in a SQLite3 database on
disk, and executes queries to bring only the required information into memory
when requested. There is some I/O overhead involved in each look-up, but
the index information persists between runs which can greatly improve start-up
time.

"""

from .memory import MemoryIndex
from .sql import SQLIndex
from .base import IndexBase

__all__ = [
    'IndexBase',
    'MemoryIndex',
    'SQLIndex'
]


"""Example script to write out the first n spectra from a library"""
import click

from mzspeclib import SpectrumLibrary
from mzspeclib.spectrum_library import FormatInferenceFailure
from mzspeclib.backends import TextSpectralLibraryWriter
from mzspeclib.cluster import SpectrumCluster
from mzspeclib.index import MemoryIndex, SQLIndex
from mzspeclib.spectrum import Spectrum

@click.command('first_n_entries')
@click.argument('inpath', type=click.Path(exists=True))
@click.option("-i", "--input-format", type=click.Choice(sorted(SpectrumLibrary.supported_file_extensions())),
              default=None)
@click.option("-n", '--spectra-to-read', type=int, default=20)
def main(inpath, input_format, spectra_to_read: int=20):
    """Read only the first `n` spectra from the input library and write them to STDOUT in text format"""
    if SQLIndex.exists(inpath):
        index_type = SQLIndex
    else:
        index_type = MemoryIndex
    click.echo(f"Opening {inpath}", err=True)
    try:
        library = SpectrumLibrary(filename=inpath, index_type=index_type, format=input_format, create_index=False)
    except FormatInferenceFailure as err:
        click.echo(f"{err}", err=True)
        raise click.Abort()

    stream = click.get_text_stream('stdout')
    writer = TextSpectralLibraryWriter(stream)
    writer.write_header(library)

    for i, entry in enumerate(library, 1):
        if i > spectra_to_read:
            break
        if isinstance(entry, Spectrum):
            writer.write_spectrum(entry)
        elif isinstance(entry, SpectrumCluster):
            writer.write_cluster(entry)

    writer.close()


if __name__ == "__main__":
    main.main()


from typing import Iterator, List, Optional, Union

from mzspeclib.cluster import SpectrumCluster
from mzspeclib.index.memory import MemoryIndex, IndexRecord, ClusterIndexRecord
from mzspeclib.spectrum import Spectrum

from .base import SpectralLibraryBackendBase


class InMemorySpectrumLibrary(SpectralLibraryBackendBase):
    """
    An in-memory spectrum library backend that holds all its entries
    in memory.

    This can be used when generating a library in-silico.
    """

    spectra: List[Spectrum]
    clusters: List[SpectrumCluster]
    index: MemoryIndex

    @classmethod
    def guess_from_filename(cls, filename) -> bool:
        return filename == ":memory:"

    def __init__(self, spectra: Optional[List[Spectrum]]=None, clusters: Optional[List[SpectrumCluster]]=None, create_index: bool = True, **kwargs):
        super().__init__(":memory:")
        self.spectra = spectra or []
        self.clusters = clusters or []

        if create_index:
            self.create_index()

    def read_header(self) -> bool:
        return False

    def read(self) -> Iterator[Union[Spectrum, SpectrumCluster]]:
        yield from self.clusters
        yield from self.spectra

    def get_spectrum(self, spectrum_number: Optional[int] = None, spectrum_name: Optional[str] = None) -> Spectrum:
        # keep the two branches separate for the possibility that this is not
        # possible with all index schemes.
        if spectrum_number is not None:
            if spectrum_name is not None:
                raise ValueError("Provide only one of spectrum_number or spectrum_name")
            index_record = self.index.record_for(spectrum_number)
            offset = index_record.offset
        elif spectrum_name is not None:
            index_record = self.index.record_for(spectrum_name)
            offset = index_record.offset
            spectrum_number = index_record.number
        else:
            raise ValueError("Must provide either spectrum_number or spectrum_name argument")
        spectrum = self.spectra[offset]
        return spectrum

    def get_cluster(self, cluster_number: int) -> SpectrumCluster:
        offset = self.index.offset_for_cluster(cluster_number)
        cluster = self.clusters[offset]
        return cluster

    def create_index(self) -> int:
        index = MemoryIndex()
        for i, spec in enumerate(self.spectra):
            index.add(IndexRecord(spec.key, i, spec.name, None, i, None))
        for i, clus in enumerate(self.clusters):
            index.add_cluster(
                ClusterIndexRecord(
                    clus.key, i,None
                )
            )
        index.commit()
        self.index = index
        return len(index)

    def append(self, spectrum: Spectrum):
        """Append a :class:`~.Spectrum` to the in-memory buffer"""
        i = len(self.spectra)
        spectrum.index = i
        self.spectra.append(spectrum)
        if self.index:
            self.index.add(IndexRecord(spectrum.key, i, spectrum.name, None, i, None))
            self.index.commit()

    def append_cluster(self, cluster: SpectrumCluster):
        """Append a :class:`~.SpectrumCluster` to the in-memory buffer"""
        i = len(self.clusters)
        self.clusters.append(cluster)
        if self.index:
            self.index.add_cluster(ClusterIndexRecord(cluster.key, i, None))


from collections import defaultdict

from matplotlib import pyplot as plt
from matplotlib import patheffects as path_effects

import numpy as np

from .annotation import InvalidAnnotation, Unannotated


series_to_color = {
    'b': 'crimson',
    'y': 'steelblue',
    'a': 'darkorange',
    'c': 'violet',
    'z': 'dodgerblue',
    'internal': 'purple',
    'precursor': 'goldenrod',
    'immonium': 'maroon',
    'reporter': 'skyblue',
    'external': 'grey',
    'formula': 'forestgreen'
}


def peaklist_to_vector(peaklist, width=0.000001):
    """
    Convert a list of discrete centroided peaks into a pair of continuous m/z
    and intensity arrays

    Parameters
    ----------
    peaklist : :class:`~Iterable` of :class:`~.PeakLike`
        The collection of peaks to convert
    width : float, optional
        The spacing between the center of the peak and it's shoulders

    Returns
    -------
    np.ndarray:
        The generated m/z array
    np.ndarray:
        The generated intensity array

    Raises
    ------
    TypeError
        When the input could not be coerced into a peak list
    """
    mzs = []
    intensities = []
    for peak in sorted(peaklist, key=lambda x: x[0]):
        mzs.append(peak[0] - width)
        intensities.append(0.)
        mzs.append(peak[0])
        intensities.append(peak[1])
        mzs.append(peak[0] + width)
        intensities.append(0.)
    return np.array(mzs), np.array(intensities)


def draw_spectrum(spectrum, ax=None, normalize=False, label_threshold=0.1, label_rotation=0, **kwargs):
    """
    Draw and annotate a Spectrum.

    Parameters
    ----------
    spectrum: :class:`~.Spectrum`
        The annotated spectrum to draw.
    ax : matplotlib.Axes, optional
        The axis to draw the plot on. If missing, a new one will be created using
        :func:`matplotlib.pyplot.subplots`
    normalize: bool, optional
        if `True`, will normalize the abundance dimension to be between 0 and 100%
    pretty: bool, optional
        If `True`, will call :func:`_beautify_axes` on `ax`
    label_threshold: float, optional
        The minimum intensity to label a peak for
    label_rotation : float, optional
        The degrees of rotation to apply to peak labels
    **kwargs
        Passed to :meth:`matplotlib.Axes.plot`

    Returns
    -------
    matplotlib.Axes
    """
    pretty = kwargs.pop("pretty", True)
    if ax is None:
        _, ax = plt.subplots(1)
    mz_array, intensity_array = peaklist_to_vector(spectrum.peak_list)
    unscaled_max = intensity_array.max()

    if normalize:
        intensity_array = intensity_array / intensity_array.max() * 100.0

    kwargs.setdefault("color", 'black')
    kwargs.setdefault("lw", 0.75)
    ax.plot(mz_array, intensity_array, **kwargs)

    max_intensity = intensity_array.max()
    ypad = 0.01 * max_intensity
    threshold = label_threshold * max_intensity
    by_series = defaultdict(list)
    for peak in spectrum.peak_list:
        height = peak[1]
        if normalize:
            height /= unscaled_max / 100.0
        if threshold >= height:
            continue
        height += ypad
        if peak[2] and not isinstance(peak[2][0], (Unannotated, )):
            annots = peak[2]
            for annot in annots:
                if isinstance(annot, InvalidAnnotation):
                    continue
                if annot.series_label == 'peptide':
                    by_series[annot.series].append(peak)
                else:
                    by_series[annot.series_label].append(peak)
            txt = ax.text(peak[0], height,
                          ',\n'.join(filter(lambda x: x != '?', map(str, annots))),
                          ha='center', rotation=label_rotation,
                          clip_on=True)
            txt.set_path_effects([path_effects.Stroke(linewidth=0.75, foreground='white'),
                                path_effects.Normal()])

    for annot_series, peaks in by_series.items():
        sub_mz_array, sub_intensity_array = peaklist_to_vector(peaks)

        if normalize:
            sub_intensity_array = sub_intensity_array / sub_intensity_array.max() * 100.0
        ax.plot(
            sub_mz_array,
            sub_intensity_array,
            label=annot_series,
            color=series_to_color[annot_series],
            lw=2.0
        )


    ax.set_xlabel("m/z")
    ax.set_ylabel("Relative Intensity")
    ax.set_title(f"{spectrum.name}")
    if pretty:
        if intensity_array.shape[0] > 0:
            set_ylim = intensity_array.min() >= 0
        else:
            set_ylim = True
        _beautify_axes(ax, set_ylim, normalize)
    return ax


def _beautify_axes(ax, set_ylim=True, normalize=False):
    ax.axes.spines['right'].set_visible(False)
    ax.axes.spines['top'].set_visible(False)
    ax.yaxis.tick_left()
    ax.xaxis.tick_bottom()
    ax.xaxis.set_ticks_position('none')
    ax.get_xaxis().get_major_formatter().set_useOffset(False)
    if set_ylim:
        ax.set_ylim(0, max(ax.get_ylim()))
    if normalize:
        _normalize_ylabels(ax)
        # ax.set_ylim(0, max(ax.get_ylim()) * 1.15)
    xlim = ax.get_xlim()
    ax.set_xlim(max(xlim[0] - 100, 0), xlim[1] + 50)
    return ax


def _normalize_ylabels(ax, adjust_sign=True):
    ytick_labels = ax.get_yticklabels()
    yticks = ax.get_yticks()
    for tl, tick in zip(ytick_labels, yticks):
        txt = tl.get_text()
        if txt == "":
            txt = str(tick)
        if adjust_sign:
            txt = txt.replace("-", "")
        txt += "%"
        tl.set_text(txt)
    ax.set_yticklabels(ytick_labels)
    return ax



"""
File Format Backends
--------------------

"""

from .text import TextSpectralLibrary, TextSpectralLibraryWriter
from .json import JSONSpectralLibrary, JSONSpectralLibraryWriter
from .msp import MSPSpectralLibrary, MSPSpectralLibraryWriter
from .bibliospec import BibliospecSpectralLibrary, BlibIndex
from .sptxt import SPTXTSpectralLibrary
from .diann import DiaNNTSVSpectralLibrary, DIANNTSVSpectralLibrary
from .spectronaut import SpectronautTSVSpectralLibrary
from .encyclopedia import EncyclopediaSpectralLibrary, EncyclopediaIndex
from .memory import InMemorySpectrumLibrary
from .base import (
    guess_implementation,
    SpectralLibraryBackendBase,
    SpectralLibraryWriterBase,
    FormatInferenceFailure,
    AttributeSetTypes,
    LibraryIterator,
)



from dataclasses import dataclass
import io
import re

from typing import Any, Dict, List, Tuple

from mzpaf.annotation import IonAnnotationBase, tokenize_signed_symbol_list

from mzspeclib.analyte import Analyte, Interpretation, InterpretationMember
from mzspeclib.attributes import AttributeManager, Attributed

from mzspeclib.annotation import AnnotationStringParser
from mzspeclib.spectrum import Spectrum

from .utils import open_stream, CaseInsensitiveDict, try_cast
from .base import DEFAULT_VERSION, FORMAT_VERSION, LIBRARY_NAME
from .msp import (
    DispatchingAttributeHandler,
    MSPSpectralLibrary as _MSPSpectralLibrary,
    analyte_terms,
    AttributeRulesEngine,
    AttributeRuleBase,
    AttributeHandler,
    MappingAttributeHandler,
    FunctionAttributeHandler,
    PeakParsingStrategy,
    FullNameParsingRule,
    StrippedPeptideParserRule,
    collision_energy_handler,
    nreps_handler,
    mz_diff_handler,
    protein_handler as msp_protein_handler,
    PeptideNotationUnpackingRule,
)

ANNOTATION_PATTERN = re.compile(
    r"""^(?P<is_auxiliary>&)?
(?:(?:(?P<series>[abcxyz]\.?)(?P<ordinal>\d+))|
   (:?Int/(?P<series_internal>[ARNDCEQGHKMFPSTWYVILJarndceqghkmfpstwyvilj]+))|
   (?P<precursor>p)|
   (:?I(?P<immonium>[ARNDCEQGHKMFPSTWYVIL]?)A?(?:(?P<immonium_modification>CAM)|[A-Z])?)|
   (?:_(?P<external_ion>[^\s,/]+))
)
(?P<neutral_losses>[+-]\d*(?:\.\d+)?)?
(?:\^(?P<charge>[+-]?\d+))?
(?:\[M(?P<adducts>(:?[+-]\d*[A-Z][A-Za-z0-9]*)+)\])?
(?:(?P<isotope>\d*)i)?
(?:/(?P<mass_error>[+-]?\d+(?:\.\d+))(?P<mass_error_unit>ppm)?)?
""",
    re.X,
)


HEADER_MAX_DEPTH = 1000


class SPTXTAnnotationParser(AnnotationStringParser):
    nominal_loss_to_formula = {
        "18": ["H2O"],
        "17": ["NH3"],
        "28": ["CO"],
        "34": ["NH3", "NH3"],
        "35": ["H2O", "NH3"],
        "36": ["H2O", "H2O"],
        "44": ["CO2"],
        "43": ["HNCO"],
        "45": ["HCONH2"],
        "46": ["HCOOH"],
        "64": ["CH4OS"],
        "82": ["CH4OS", "H2O"],
        "91": ["C2H5NOS"],
        "92": ["C2H4O2S"],
    }

    def __init__(self, pattern):
        super().__init__(pattern)

    def _dispatch(
        self,
        annotation_string,
        data,
        adducts,
        charge,
        isotope,
        neutral_losses,
        analyte_reference,
        mass_error,
        confidence,
        **kwargs,
    ):
        data["sequence_ordinal"] = None
        return super()._dispatch(
            annotation_string,
            data,
            adducts,
            charge,
            isotope,
            neutral_losses,
            analyte_reference,
            mass_error,
            confidence,
            **kwargs,
        )

    def parse_annotation(self, annotation_string: str, **kwargs) -> List[IonAnnotationBase]:
        parts = annotation_string.split(",")
        patched_token = []
        for part in parts:
            if part.startswith("[") and part.endswith("]"):
                patched_token.append("&" + part[1:-1])
            else:
                patched_token.append(part)
        annotation_string = ','.join(patched_token)
        return super().parse_annotation(annotation_string, **kwargs)

    def _convert_losses(self, neutral_losses: str):
        tokens = []
        for i, loss in enumerate(tokenize_signed_symbol_list(neutral_losses)):
            if loss[0] == '-':
                sign = '-'
                loss = loss[1:]
            else:
                sign = '+'
            if loss in self.nominal_loss_to_formula:
                loss = sign.join(self.nominal_loss_to_formula[loss])
            tokens.append(sign)
            tokens.append(loss)
        return ''.join(tokens)


    def _parse_string(self, annotation_string: str, **kwargs) -> Tuple[re.Match, Dict[str, str]]:
        match, data = super()._parse_string(annotation_string, **kwargs)
        if data["isotope"] is not None and not data["isotope"]:
            data["isotope"] = "1"
        if data['neutral_losses']:
            try:
                data["neutral_losses"] = self._convert_losses(data['neutral_losses'])
            except KeyError as err:
                print(annotation_string)
                raise err from None
        return match, data


parse_annotation = SPTXTAnnotationParser(ANNOTATION_PATTERN)

@dataclass
class FullNameParserSPTXT(FullNameParsingRule):
    attribute = "FullName"

    name_pattern = re.compile(r"([A-Z\-\*])\.((?:[A-Z](?:\[[^\]]+\])?)+)\.([A-Z\-\*])/*([\d]*)")
    tokenizer = re.compile(r"([A-Z])(\[[^\]]+?\])?")

    def parse_value(self, value: str):
        match = self.name_pattern.match(value)
        if match is None:
            return None
        peptide = match.group(2)
        nterm = match.group(1)
        cterm = match.group(3)
        charge = match.group(4)

        peptide = ''.join([aa for aa, mod in self.tokenizer.findall(peptide)])
        return peptide, nterm, cterm, charge


def tpp_spectrum_to_usi(spectrum: str) -> str:
    tokens = spectrum.split(".")
    source_file = tokens[0]
    scan_number = int(tokens[1])
    scan_number2 = tokens[2]
    return (f"mzspec:USI000000:{source_file}:scan:{scan_number}", source_file, scan_number)


@FunctionAttributeHandler.wraps("RawSpectrum", "BestRawSpectrum")
def raw_spectrum_handler(key, value, container: Attributed):
    if key == "RawSpectrum":
        usi, source_file, scan_number = tpp_spectrum_to_usi(value)
        container.add_attribute("MS:1003203|constituent spectrum file", source_file)
        container.add_attribute("MS:1003299|contributing replicate spectrum USI", usi)
        return True
    elif key == "BestRawSpectrum":
        usi, source_file, scan_number = tpp_spectrum_to_usi(value)
        container.add_attribute("MS:1003203|constituent spectrum file", source_file)
        container.add_attribute("MS:1003299|contributing replicate spectrum USI", usi)
        return True
    else:
        return False


@FunctionAttributeHandler.wraps("RetentionTime")
def retention_time_handler(key: str, value: Any, container: Attributed):
    rt_rep, rt_min, rt_max = map(float, value.split(","))
    group_identifier = container.get_next_group_identifier()
    container.add_attribute(
        "MS:1000894|retention time",
        rt_rep,
        group_identifier,
    )
    container.add_attribute("UO:0000000|unit", "UO:0000010|second", group_identifier)
    return True


_HCD = [
    "MS:1000044|dissociation method",
    "MS:1000422|beam-type collision-induced dissociation",
]
_CID = [
    "MS:1000044|dissociation method",
    "MS:1002472|trap-type collision-induced dissociation",
]
_CAD = ["MS:1000044|dissociation method", "MS:1000133|collision-induced dissociation"]

INSTRUMENT_DISPATCH = CaseInsensitiveDict(
    {
        "it": [_CID],
        "hcd": [_HCD],
        "qtof": [_CAD],
        "qqq": [_CAD],
        "qit": [_CAD],
    }
)


@FunctionAttributeHandler.wraps("Inst")
def instrument_handler(key, value, container: Attributed):
    value = str(value)
    if "/" in value:
        _inst_id, rest = value.split("/", 1)
        inst_type, rest = rest.split(",", 1)
        for k, v in INSTRUMENT_DISPATCH[inst_type]:
            container.add_attribute(k, v)
        return True
    else:
        container.add_attribute(*_HCD)
        return True


INTERPRETATION_TERMS = CaseInsensitiveDict({"FracUnassigned": "MS:1003079|total unassigned intensity fraction"})


@FunctionAttributeHandler.wraps("DeltaCn", "DotConsensus", "Se", "Prob")
def search_engine_keys(key: str, value: Any, container: Attributed):
    container.add_attribute_group(
        [
            ["MS:1003275|other attribute name", key],
            ["MS:1003276|other attribute value", value],
        ]
    )
    return True


INTERPRETATION_MEMBER_TERMS = DispatchingAttributeHandler(
    [MappingAttributeHandler(CaseInsensitiveDict({"XCorr": "MS:1002252|Comet:xcorr"})), (search_engine_keys)]
)

OTHER_TERMS = CaseInsensitiveDict(
    {
        "Single": [
            "MS:1003065|spectrum aggregation type",
            "MS:1003066|singleton spectrum",
        ],
        "Consensus": [
            "MS:1003065|spectrum aggregation type",
            "MS:1003067|consensus spectrum",
        ],
        "Inst": instrument_handler,
        "Spec": {
            "Consensus": [
                [
                    "MS:1003065|spectrum aggregation type",
                    "MS:1003067|consensus spectrum",
                ]
            ],
            "Single": [
                [
                    "MS:1003065|spectrum aggregation type",
                    "MS:1003066|singleton spectrum",
                ]
            ],
        },
        "Scan": "MS:1003057|scan number",
        "Origfile": "MS:1003203|constituent spectrum file",
        "filename": "MS:1003203|constituent spectrum file",
        "file_name": "MS:1003203|constituent spectrum file",
        "Sample": "MS:1000002|sample name",
        "Filter": "MS:1000512|filter string",
        "FTResolution": "MS:1000028|detector resolution",
        "PrecursorIntensity": "MS:1003085|previous MS1 scan precursor intensity",
        "OrigMaxIntensity": "MS:1003086|precursor apex intensity",
        "Purity": "MS:1009013|isolation window precursor purity",
        "NumPeaks": "MS:1003059|number of peaks",
        "Run": "MS:1003203|constituent spectrum file",
        "CollisionEnergy": collision_energy_handler,
    }
)


SPTXT_SPECTRUM_ATTRIBUTE_MAP = CaseInsensitiveDict(
    {
        "TotalIonCurrent": "MS:1000285|total ion current",
        "RetentionTime": retention_time_handler,
    }
)


@FunctionAttributeHandler.wraps("Protein")
def protein_handler(key, value: str, container):
    if "/" in value:
        pid, rest = value.split("/", 1)
        if pid.isdigit():
            return msp_protein_handler(key, rest, container)
        else:
            return msp_protein_handler(key, value, container)
    else:
        return msp_protein_handler(key, value, container)


SPTXT_ANALYTE_ATTRIBUTE_MAP = CaseInsensitiveDict(
    {
        "MassDiff": mz_diff_handler,
        "Protein": protein_handler,
    }
)


SPTXT_SPECTRUM_ATTRIBUTE_HANDLER = DispatchingAttributeHandler()

SPTXT_SPECTRUM_ATTRIBUTE_HANDLER.add(nreps_handler)
SPTXT_SPECTRUM_ATTRIBUTE_HANDLER.add(MappingAttributeHandler(SPTXT_SPECTRUM_ATTRIBUTE_MAP))
SPTXT_SPECTRUM_ATTRIBUTE_HANDLER.add(raw_spectrum_handler)


@dataclass
class SkippedAttributeRule(AttributeRuleBase):
    attribute: str

    def is_applicable(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ) -> bool:
        return self.attribute in attributes

    def process_attributes(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ) -> List[str]:
        attributes.pop(self.attribute)
        return []


@dataclass
class LibIDAttributeRule(AttributeRuleBase):
    attribute: str

    def is_applicable(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ) -> bool:
        return self.attribute in attributes

    def process_attributes(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ) -> List[str]:
        spectrum.key = attributes.pop(self.attribute)
        return []


class SPTXTSpectralLibrary(_MSPSpectralLibrary):
    file_format = "sptxt"
    format_name = "sptxt"

    @classmethod
    def guess_from_header(cls, filename: str) -> bool:
        with open_stream(filename, "r") as stream:
            first_line = stream.readline()
            i = 1
            while first_line.startswith("###") and i < HEADER_MAX_DEPTH:
                first_line = stream.readline()
                i += 1
            if re.match("Name: ", first_line):
                return True
            if first_line.startswith("###"):
                return True
        return False

    def read_header(self) -> bool:
        with open_stream(self.filename, "rb") as stream:
            match, offset = self._parse_header_from_stream(stream)
            return match
        return False

    def _parse_header_from_stream(self, stream: io.RawIOBase) -> Tuple[bool, int]:
        nbytes = 0
        attributes = AttributeManager()
        attributes.add_attribute(FORMAT_VERSION, DEFAULT_VERSION)

        line = stream.readline()
        nbytes += len(line)
        i = 0
        header_buffer = []
        while not re.match(b"Name:", line) and i < HEADER_MAX_DEPTH:
            header_buffer.append(line)
            line = stream.readline()
            nbytes += len(line)
            i += 1

        self._parse_header_into_attributes(header_buffer, attributes)
        if not attributes.has_attribute(LIBRARY_NAME):
            name = self._infer_lib_name()
            if name:
                attributes.add_attribute(LIBRARY_NAME, name)

        self.attributes.clear()
        self.attributes._from_iterable(attributes)
        if re.match(b"Name:", line):
            header_buffer.append(line)

            return True, nbytes - len(line)
        return False, 0

    def _parse_header_into_attributes(self, lines: List[bytes], attributes: AttributeManager):
        for i, line in enumerate(lines):
            if line.startswith(b"###"):
                if i == 0:
                    name = line.rsplit(b'(', 1)[0].strip().decode('utf8').strip("# ")
                    attributes.add_attribute(LIBRARY_NAME, name)
                elif i == 1:
                    version = line.split(b"(", 1)[1].strip().decode("utf8").split(",")[0].split(" ")[1]
                    gid = attributes.get_next_group_identifier()
                    attributes.add_attribute("MS:1003207|library creation software", "MS:1001477|SpectraST", gid)
                    attributes.add_attribute("MS:1003200|software version", version, gid)
            else:
                continue

    def _make_peak_parsing_strategy(self) -> PeakParsingStrategy:
        return PeakParsingStrategy(parse_annotation)

    def _make_attribute_handlers(self) -> AttributeRulesEngine:
        return AttributeRulesEngine(
            MappingAttributeHandler(OTHER_TERMS),
            MappingAttributeHandler(analyte_terms),
            MappingAttributeHandler(INTERPRETATION_TERMS),
            INTERPRETATION_MEMBER_TERMS,
            SPTXT_SPECTRUM_ATTRIBUTE_HANDLER,
            MappingAttributeHandler(SPTXT_ANALYTE_ATTRIBUTE_MAP),
            early_rules=[
                FullNameParserSPTXT(),
                FullNameParserSPTXT("Fullname"),
                LibIDAttributeRule("LibID"),
                SkippedAttributeRule("BinaryFileOffset"),
                # Redundant with the Pep parsing rule
                SkippedAttributeRule("NTT"),
            ],
            late_rules=[StrippedPeptideParserRule(), PeptideNotationUnpackingRule(self.modification_parser)],
        )



"""A collection of constants"""

PROTON = 1.00727646677

FIRST_ANALYTE_KEY = "1"
FIRST_INTERPRETATION_KEY = "1"


LIBRARY_CREATION_SW = "MS:1003207|library creation software"
SW_VERSION = "MS:1003200|software version"

ATTRIBUTE_SET_NAME = "MS:1003212|library attribute set name"
PEAK_ATTRIBUTE = "MS:1003254|peak attribute"

ANALYTE_MIXTURE = "MS:1003163|analyte mixture members"
CHARGE_STATE = "MS:1000041|charge state"
PROFORMA_ION = "MS:1003270|proforma peptidoform ion notation"
PROFORMA_SEQ = "MS:1003169|proforma peptidoform sequence"
STRIPPED_PEPTIDE_SEQ = "MS:1000888|stripped peptide sequence"
MOLECULAR_FORMULA = "MS:1000866|molecular formula"
RETENTION_TIME = "MS:1000894|retention time"

MOLECULAR_MASS = "MS:1000224|molecular mass"
ADDUCT_ION_MASS = "MS:1003243|adduct ion mass"
THEORETICAL_MASS = "MS:1001117|theoretical mass"

FORMAT_VERSION = "MS:1003186|library format version"
LIBRARY_NAME = "MS:1003188|library name"
LIBRARY_VERSION = "MS:1003190|library version"
LIBRARY_IDENTIFIER = "MS:1003187|library identifier"
LIBRARY_DESCRIPTION = "MS:1003189|library description"
LIBRARY_URI = "MS:1003191|library URI"

SAMPLE_NAME = "MS:1000002|sample name"
SOURCE_FILE = "MS:1003203|constituent spectrum file"
SCAN_NUMBER = "MS:1003057|scan number"
MS_LEVEL = "MS:1000511|ms level"
FILTER_STRING = "MS:1000512|filter string"

SPECTRUM_NAME = "MS:1003061|library spectrum name"
LIBRARY_SPECTRUM_KEY = "MS:1003237|library spectrum key"
LIBRARY_SPECTRUM_INDEX = "MS:1003062|library spectrum index"

SELECTED_ION_MZ = "MS:1000744|selected ion m/z"
PRECURSOR_MZ = "MS:1003208|experimental precursor monoisotopic m/z"
THEORETICAL_MZ = "MS:1003053|theoretical monoisotopic m/z"
DELTA_MZ = "MS:1001975|delta m/z"
ISOLATION_WINDOW_PRECURSOR_PURITY = "MS:1009013|isolation window precursor purity"
PRECURSOR_APEX_INTENSITY = "MS:1003086|precursor apex intensity"
BASE_PEAK_INTENSITY = "MS:1000505|base peak intensity"

SPECTRUM_ORIGIN_TYPE = "MS:1003072|spectrum origin type"
SELECTED_FRAGMENT_THEORETICAL_OBSERVED_INTENSITY_SPECTRUM = "MS:1003424|selected fragment theoretical m/z observed intensity spectrum"

SPECTRUM_AGGREGATION_TYPE = "MS:1003065|spectrum aggregation type"
SINGLETON_SPECTRUM = "MS:1003066|singleton spectrum"
CONSENSUS_SPECTRUM = "MS:1003067|consensus spectrum"
NUMBER_OF_REPLICATE_SPECTRA_USED = "MS:1003070|number of replicate spectra used"
NUMBER_OF_REPLICATE_SPECTRA_AVAILABLE = "MS:1003069|number of replicate spectra available"
DECOY_SPECTRUM = "MS:1003192|decoy spectrum"
DECOY_PEPTIDE_SPECTRUM = "MS:1003195|unnatural peptidoform decoy spectrum"

SCAN_POLARITY = "MS:1000465|scan polarity"
POSITIVE_SCAN = "MS:1000130|positive scan"
NEGATIVE_SCAN = "MS:1000129|negative scan"

CUSTOM_ATTRIBUTE_NAME = "MS:1003275|other attribute name"
CUSTOM_ATTRIBUTE_VALUE = "MS:1003276|other attribute value"

PEAK_OBSERVATION_FREQ = "MS:1003279|observation frequency of peak"
PEAK_ATTRIB = "MS:1003254|peak attribute"

DISSOCIATION_METHOD = "MS:1000044|dissociation method"
HCD = "MS:1000422|beam-type collision-induced dissociation"
TRAP_CID = "MS:1002472|trap-type collision-induced dissociation"
COLLISION_ENERGY = "MS:1000045|collision energy"

UNIT = "UO:0000000|unit"
ELECTRONVOLT = "UO:0000266|electronvolt"
PERCENT = "UO:0000187|percent"
SECOND = "UO:0000010|second"
MINUTE = "UO:0000031|minute"
MZ = "MS:1000040|m/z"
PPM = "UO:0000169|parts per million"


INTENSITY_OF_HIGH_UNASSIGNED_PEAK = "MS:1003289|intensity of highest unassigned peak"
TOP_20_UNASSIGNED_INTENSITY_FRACTION = "MS:1003080|top 20 peak unassigned intensity fraction"
TOTAL_UNASSIGNED_INTENSITY_FRACTION = "MS:1003079|total unassigned intensity fraction"
NUM_UNASSIGNED_PEAKS_IN_TOP_20 = "MS:1003290|number of unassigned peaks among top 20 peaks"
NUM_PEAKS = "MS:1003059|number of peaks"

Q_VALUE = "MS:1002354|PSM-level q-value"

PROTEIN_ACCESSION = "MS:1000885|protein accession"
PROTEIN_NAME = "MS:1000886|protein name"
PROTEIN_DESCRIPTION = "MS:1001088|protein description"

TAXONOMY_NCBI_TAX_ID = "MS:1001467|taxonomy: NCBI TaxID"
TAXONOMY_SCIENTIFIC_NAME = "MS:1001469|taxonomy: scientific name"
TAXONOMY_COMMON_NAME = "MS:1001468|taxonomy: common name"


import os

from mzspeclib import SpectrumLibrary
from mzspeclib import annotation

workspace = os.path.dirname(__file__)

source = SpectrumLibrary(filename=os.path.join(
    workspace, "chinese_hamster_hcd_selected_head.msp"))
print(source)
annots = set()
for spec in source:
    print(spec.name)
    for peak in spec.peak_list:
        for a in peak[2]:
            if isinstance(a, annotation.InvalidAnnotation):
                breakpoint()
            annots.add(a)

with open(os.path.join(workspace, "annotations.txt"), 'wt') as fh:
    fh.write("\n".join(sorted(map(str, annots))))



import math
import os
import unittest

from mzspeclib.spectrum import Spectrum
from mzspeclib.backends import (MSPSpectralLibrary, TextSpectralLibrary, JSONSpectralLibrary, SpectronautTSVSpectralLibrary, DIANNTSVSpectralLibrary)
from mzspeclib.analyte import ANALYTE_MIXTURE

from .common import datafile


class LibraryBehaviorBase(object):

    def _open_library(self, file_path=None):
        if file_path is None:
            file_path = self.test_file
        lib = self.library_cls(file_path)
        return lib

    def test_sequence_behavior(self):
        lib = self._open_library()
        assert len(lib) == 7
        spec = lib[3]
        assert spec.get_attribute(
            "MS:1003061|library spectrum name") == "AAAAGSTSVKPIFSR/2_0_44eV"

    # TODO: Fix clipping in _buffer_from_stream first
    def test_iteration(self):
        lib = self._open_library()
        assert list(lib) == list(lib.read())


class TestMSPLibrary(unittest.TestCase, LibraryBehaviorBase):
    test_file = datafile("chinese_hamster_hcd_selected_head.msp")
    library_cls = MSPSpectralLibrary


class MzSpecLibLibraryBehaviorBase(LibraryBehaviorBase):
    def test_interpretation_attribute(self):
        try:
            path = self.test_interpretation_file
        except AttributeError:
            return
        lib = self._open_library(path)
        spec: Spectrum = lib[0]
        interp = spec.interpretations[1]
        assert len(interp.attributes) == 2
        assert interp.get_attribute('MS:1002357|PSM-level probability') == 0.974
        assert interp.get_attribute(ANALYTE_MIXTURE) == [1, 2]
        mem = interp.get_member_interpretation(1)
        assert len(mem) == 1
        assert mem.get_attribute('MS:XXXXXXX|explained intensity fraction') == 0.287


class TestTextLibrary(unittest.TestCase, MzSpecLibLibraryBehaviorBase):
    test_file = datafile("chinese_hamster_hcd_selected_head.mzlb.txt")
    library_cls = TextSpectralLibrary
    test_interpretation_file = datafile("complex_interpretations_with_members.mzlb.txt")


class TestJSONLibrary(unittest.TestCase, MzSpecLibLibraryBehaviorBase):
    test_file = datafile("chinese_hamster_hcd_selected_head.mzlb.json")
    library_cls = JSONSpectralLibrary
    test_interpretation_file = datafile("complex_interpretations_with_members.mzlb.json")


class TestSpectronautLibrary(unittest.TestCase, LibraryBehaviorBase):
    test_file = datafile("human_serum.head.spectronaut.tsv")
    library_cls = SpectronautTSVSpectralLibrary

    def test_sequence_behavior(self):
        lib = self._open_library()
        assert len(lib) == 9

        spec: Spectrum = lib[0]
        assert spec.name == 'AQIPILR/2'
        assert math.isclose(spec.precursor_mz, 405.7634379)
        assert spec.precursor_charge == 2

        spec = lib[5]
        assert spec.name == 'QELSEAEQATR/2'
        assert math.isclose(spec.precursor_mz, 631.3045807)
        assert spec.get_analyte(1).proteins[0].name == 'CO3_HUMAN'


class TestDIANNTSVLibrary(unittest.TestCase, LibraryBehaviorBase):
    test_file = datafile("phl004_canonical_sall_pv_plasma.head.diann.tsv")
    library_cls = DIANNTSVSpectralLibrary

    def test_sequence_behavior(self):
        lib = self._open_library()
        assert len(lib) == 9

        spec: Spectrum = lib[0]
        analyte = spec.get_analyte(1)
        assert analyte.peptide == 'AAAAAAAAAAAAAAAASAGGK/2'
        assert spec.name == 'AAAAAAAAAAAAAAAASAGGK2'

